from typing import List, Optional

from alitra import Pose
from pydantic import BaseModel, Field

from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import TASKS, TaskTypes
from robot_interface.utilities.uuid_string_factory import uuid4_string


class Mission(BaseModel):
    id: str = Field(default_factory=uuid4_string, frozen=True)
    tasks: List[TASKS] = Field(default_factory=list, frozen=True)
    name: str = Field(frozen=True)
    start_pose: Optional[Pose] = Field(default=None, frozen=True)
    status: MissionStatus = MissionStatus.NotStarted
    error_message: Optional[ErrorMessage] = Field(default=None)

    def _is_return_to_home_mission(self) -> bool:
        if len(self.tasks) != 1:
            return False
        if self.tasks[0].type != TaskTypes.ReturnToHome:
            return False
        return True



import time
from collections import deque
from http import HTTPStatus

import sqlalchemy
from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture
from sqlalchemy.orm import Session
from testcontainers.mysql import MySqlContainer

from isar.services.service_connections.persistent_memory import (
    Base,
    PersistentRobotState,
    change_persistent_robot_state_is_maintenance_mode,
    read_persistent_robot_state_is_maintenance_mode,
)
from isar.state_machine.states_enum import States
from robot_interface.models.mission.status import MissionStatus, RobotStatus
from tests.test_mocks.mission_definition import DummyMissionDefinition
from tests.test_mocks.robot_interface import StubRobot
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
)


def test_persistent_storage_schema() -> None:
    with MySqlContainer("mysql:9.4.0", dialect="pymysql") as mysql:
        connection_url = mysql.get_connection_url()
        engine = sqlalchemy.create_engine(connection_url)
        Base.metadata.create_all(engine)

        with Session(engine) as session:
            persistent_state = PersistentRobotState(
                robot_id="0a0", is_maintenance_mode=True
            )
            session.add_all([persistent_state])
            session.commit()

        is_maintenance_mode = read_persistent_robot_state_is_maintenance_mode(
            connection_url, "0a0"
        )
        assert is_maintenance_mode
        change_persistent_robot_state_is_maintenance_mode(
            connection_url, "0a0", value=False
        )
        is_maintenance_mode = read_persistent_robot_state_is_maintenance_mode(
            connection_url, "0a0"
        )
        assert not is_maintenance_mode


def test_maintenance_mode(
    setup_db_connection_string: str,  # The order of the fixtures is important
    client: TestClient,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    # Now running ISAR should put it into maintenance mode
    state_machine_thread.start()
    robot_service_thread.start()

    assert state_machine_thread.state_machine.current_state.name == States.Maintenance

    # The robot should have started in maintenance mode since the robot id is not found in the database.
    response = client.post(
        url="/schedule/start-mission",
        json=jsonable_encoder(
            {
                "mission_definition": DummyMissionDefinition.dummy_start_mission_definition
            }
        ),
    )
    assert response.status_code == HTTPStatus.CONFLICT

    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    t_start = time.time()
    while (
        state_machine_thread.state_machine.shared_state.robot_status.check()
        != RobotStatus.Home
    ):
        if time.time() - t_start > 10:
            raise Exception("Robot did not come Home within expected time")
        time.sleep(0.5)

    response = client.post(url="/schedule/release-maintenance-mode")
    assert response.status_code == HTTPStatus.OK
    assert state_machine_thread.state_machine.current_state.name == States.Home

    mocker.patch.object(
        StubRobot, "mission_status", return_value=MissionStatus.InProgress
    )  # The robot will not go to awaitng next mission after mission has started, it should remain in monitor. In order to test the "stop" functionality.
    response = client.post(
        url="/schedule/start-mission",
        json=jsonable_encoder(
            {
                "mission_definition": DummyMissionDefinition.dummy_start_mission_definition
            }
        ),
    )
    assert response.status_code == HTTPStatus.OK

    assert state_machine_thread.state_machine.current_state.name == States.Monitor
    response = client.post(url="/schedule/maintenance-mode")
    assert response.status_code == HTTPStatus.OK

    response = client.post(
        url="/schedule/start-mission",
        json=jsonable_encoder(
            {
                "mission_definition": DummyMissionDefinition.dummy_start_mission_definition
            }
        ),
    )
    assert response.status_code == HTTPStatus.CONFLICT

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Maintenance,
            States.Home,
            States.Monitor,
            States.StoppingDueToMaintenance,
            States.Maintenance,
        ]
    )



from importlib import import_module

from dependency_injector import containers, providers

from isar.apis.api import API
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.settings import settings
from isar.models.events import Events, SharedState
from isar.robot.robot import Robot
from isar.services.utilities.robot_utilities import RobotUtilities
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.storage.blob_storage import BlobStorage
from isar.storage.local_storage import LocalStorage
from isar.storage.uploader import Uploader
from robot_interface.telemetry.mqtt_client import MqttPublisher


class ApplicationContainer(containers.DeclarativeContainer):
    config = providers.Configuration(pydantic_settings=[settings])

    # Events and shared state
    events = providers.Singleton(Events)
    shared_state = providers.Singleton(SharedState)

    # Robot-related services
    robot_interface = providers.Singleton(
        lambda: import_module(f"{settings.ROBOT_PACKAGE}.robotinterface").Robot()
    )
    robot_utilities = providers.Singleton(RobotUtilities, robot=robot_interface)

    # Mqtt client
    mqtt_client = providers.Singleton(
        MqttPublisher,
        mqtt_queue=providers.Callable(events.provided.mqtt_queue),
    )

    # API and controllers
    authenticator = providers.Singleton(Authenticator)
    scheduling_utilities = providers.Singleton(
        SchedulingUtilities,
        events=events,
        shared_state=shared_state,
    )
    scheduling_controller = providers.Singleton(
        SchedulingController, scheduling_utilities=scheduling_utilities
    )
    robot_controller = providers.Singleton(
        RobotController, robot_utilities=robot_utilities
    )
    api = providers.Singleton(
        API,
        authenticator=authenticator,
        scheduling_controller=scheduling_controller,
        robot_controller=robot_controller,
        mqtt_publisher=mqtt_client,
    )

    # Storage
    storage_handlers_temp = []
    if settings.STORAGE_LOCAL_ENABLED:
        local_storage = providers.Singleton(LocalStorage)
        storage_handlers_temp.append(local_storage)
    if settings.STORAGE_BLOB_ENABLED:
        blob_storage = providers.Singleton(BlobStorage)
        storage_handlers_temp.append(blob_storage)
    storage_handlers = providers.List(*storage_handlers_temp)

    # State machine
    state_machine = providers.Singleton(
        StateMachine,
        events=events,
        shared_state=shared_state,
        robot=robot_interface,
        mqtt_publisher=mqtt_client,
    )

    # Robot
    robot = providers.Singleton(
        Robot,
        events=events,
        robot=robot_interface,
        shared_state=shared_state,
        mqtt_publisher=mqtt_client,
    )

    # Uploader
    uploader = providers.Singleton(
        Uploader,
        events=events,
        storage_handlers=storage_handlers,
        mqtt_publisher=mqtt_client,
    )


def get_injector() -> ApplicationContainer:
    container = ApplicationContainer()
    container.init_resources()
    container.wire(modules=[__name__])

    print("Loaded the following module configurations:")
    for provider_name, provider in container.providers.items():
        provider_repr = repr(provider)
        simplified_provider = provider_repr.split(".")[-1].split(">")[0]
        print(f"    {provider_name:<20}: {simplified_provider}")

    return container



import ast
import os
from typing import List, Optional

from python_to_mermaid import MermaidDiagram


def get_imports(source_code: str) -> tuple[Optional[str], list]:
    own_class_name = None
    imported_states = []
    tree = ast.parse(source_code)
    for node in ast.walk(tree):
        if isinstance(node, ast.Import) and node.names[0].name.startswith(
            "isar.state_machine.states."
        ):
            state_name = node.names[0].asname
            imported_states.append(state_name)
        if isinstance(node, ast.ClassDef):
            own_class_name = node.name

    return own_class_name, imported_states


def get_all_state_file_paths() -> List[str]:
    cwd = os.getcwd()
    states_folder = os.path.join(cwd, "src", "isar", "state_machine", "states")
    state_files = []
    for file_or_folder in os.listdir(states_folder):
        sub_path = os.path.join(states_folder, file_or_folder)
        if os.path.isfile(sub_path) and not sub_path.endswith("__.py"):
            state_files.append(sub_path)
    return state_files


if __name__ == "__main__":
    state_files = get_all_state_file_paths()

    state_graph = {}

    for state_file in state_files:

        with open(state_file, "r") as file:
            source = file.read()

            own_name, imported_states = get_imports(source)
            state_graph[own_name] = imported_states

    diagram = MermaidDiagram()
    for state in state_graph:
        diagram.add_node(state)

    for state_name, transitions in state_graph.items():
        for transition in transitions:
            diagram.add_edge(state_name, transition)

    mermaid_diagram = str(diagram)

    with open("diagram.mmd", "w") as file:
        file.write(mermaid_diagram)



from typing import TYPE_CHECKING, List, Optional, Union

import isar.state_machine.states.await_next_mission as AwaitNextMission
import isar.state_machine.states.blocked_protective_stop as BlockedProtectiveStop
import isar.state_machine.states.lockdown as Lockdown
import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.offline as Offline
import isar.state_machine.states.recharging as Recharging
import isar.state_machine.states.returning_home as ReturningHome
import isar.state_machine.states.stopping as Stopping
import isar.state_machine.states.unknown_status as UnknownStatus
from isar.apis.models.models import (
    LockdownResponse,
    MaintenanceResponse,
    MissionStartResponse,
)
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Home(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events
        shared_state = state_machine.shared_state

        # This clears the current robot status value, so we don't read an outdated value
        events.robot_service_events.robot_status_changed.clear_event()

        def _send_to_lockdown_event_handler(
            should_send_robot_home: EmptyMessage,
        ) -> Transition[Lockdown.Lockdown]:
            events.api_requests.send_to_lockdown.response.trigger_event(
                LockdownResponse(lockdown_started=True)
            )
            return Lockdown.transition()

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Optional[Transition[Maintenance.Maintenance]]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        def _robot_status_event_handler(
            has_changed: EmptyMessage,
        ) -> Optional[
            Union[
                Transition[AwaitNextMission.AwaitNextMission],
                Transition[Offline.Offline],
                Transition[BlockedProtectiveStop.BlockedProtectiveStop],
                Transition[UnknownStatus.UnknownStatus],
            ]
        ]:
            robot_status: Optional[RobotStatus] = shared_state.robot_status.check()
            if robot_status == RobotStatus.Home:
                return None
            elif robot_status == RobotStatus.Available:
                self.logger.info(
                    "Got robot status available while in home state. Leaving home state."
                )
                return AwaitNextMission.transition()
            elif robot_status == RobotStatus.Offline:
                self.logger.info(
                    "Got robot status offline while in home state. Leaving home state."
                )
                return Offline.transition()
            elif robot_status == RobotStatus.BlockedProtectiveStop:
                self.logger.info(
                    "Got robot status blocked protective stop while in home state. Leaving home state."
                )
                return BlockedProtectiveStop.transition()
            self.logger.info(
                f"Got unexpected status {robot_status} while in home state. Leaving home state."
            )
            return UnknownStatus.transition()

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[Recharging.Recharging]]:
            if battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD:
                return None

            return Recharging.transition()

        def _stop_mission_event_handler(
            mission_id: str,
        ) -> Transition[Stopping.Stopping]:
            return Stopping.transition_and_trigger_stop(mission_id, True)

        def _start_mission_event_handler(
            mission: Mission,
        ) -> Optional[Transition[Monitor.Monitor]]:
            if not state_machine.battery_level_is_above_mission_start_threshold():
                events.api_requests.start_mission.response.trigger_event(
                    MissionStartResponse(
                        mission_id=mission.id,
                        mission_started=False,
                        mission_not_started_reason="Robot battery too low",
                    )
                )
                return None
            return Monitor.transition_and_start_mission(mission, True)

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[Mission](
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=_start_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="return_home_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: ReturningHome.transition_and_start_mission(True),
            ),
            EventHandlerMapping[str](
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=_stop_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="robot_status_event",
                event=events.robot_service_events.robot_status_changed,
                handler=_robot_status_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Home,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition["Home"]:
    def _transition(state_machine: "StateMachine") -> Home:
        return Home(state_machine)

    return _transition



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.stopping_return_home import StoppingReturnHome
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from tests.test_mocks.task import StubTask


def test_return_home_cancelled_when_new_mission_received(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("start_mission_event")
    )

    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])

    assert event_handler is not None

    transition = event_handler.handler(mission)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is StoppingReturnHome


def test_stopping_return_home_mission_fails(
    sync_state_machine: StateMachine,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    sync_state_machine.current_state = StoppingReturnHome(sync_state_machine, mission)
    stopping_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_return_home_state.get_event_handler_by_name("failed_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(error_description="", error_reason=ErrorReason.RobotAPIException)
    )

    assert sync_state_machine.events.api_requests.start_mission.response.has_event()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturningHome


def test_stopping_return_home_mission_succeeds(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(90.0)
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    sync_state_machine.current_state = StoppingReturnHome(sync_state_machine, mission)
    stopping_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_return_home_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.going_to_lockdown import GoingToLockdown
from isar.state_machine.states.lockdown import Lockdown
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.stopping_go_to_lockdown import StoppingGoToLockdown


def test_mission_stopped_when_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = Monitor(sync_state_machine, "mission_id")

    monitor_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        monitor_state.get_event_handler_by_name("send_to_lockdown_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is StoppingGoToLockdown


def test_going_to_lockdown_transitions_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = GoingToLockdown(sync_state_machine)

    going_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_lockdown_state.get_event_handler_by_name("mission_succeeded_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Lockdown



import logging
import logging.config
import os
from importlib.resources import as_file, files
from typing import Any

import yaml
from uvicorn.logging import ColourizedFormatter

from isar.config.settings import settings

from .settings import Settings


def setup_loggers() -> None:
    log_config = load_log_config()

    logging.config.dictConfig(log_config)

    env_log_levels = {}
    for env_var, value in os.environ.items():
        if env_var.endswith("_LOG_LEVEL"):
            log_name = env_var.split("_LOG_LEVEL")[0].lower()
            env_log_levels[log_name] = value.upper()

    handlers = []
    if settings.LOG_HANDLER_LOCAL_ENABLED:
        handlers.append(
            configure_console_handler(log_config=log_config, settings=settings)
        )

    for log_handler in handlers:
        for logger_name, logger_config in log_config["loggers"].items():
            logger = logging.getLogger(logger_name)
            logger.addHandler(log_handler)
            if "level" in logger_config:
                logger.setLevel(logger_config["level"])
            if logger_name in env_log_levels:
                logger.setLevel(env_log_levels[logger_name])
        root_logger = logging.getLogger()
        root_logger.addHandler(log_handler)
        if "level" in log_config.get("root", {}):
            root_logger.setLevel(log_config["root"]["level"])


def load_log_config() -> dict[str, Any]:
    source = files("isar").joinpath("config").joinpath("logging.conf")
    with as_file(source) as f:
        log_config = yaml.safe_load(open(f))
    return log_config


def configure_console_handler(log_config: dict, settings: Settings) -> logging.Handler:
    handler = logging.StreamHandler()
    handler.setLevel(log_config["root"]["level"])
    if settings.DEBUG_LOG_FORMATTER:
        handler.setFormatter(
            ColourizedFormatter(
                log_config["formatters"]["debug-formatter"]["format"],
                style="{",
                use_colors=True,
            )
        )
    else:
        handler.setFormatter(
            ColourizedFormatter(
                log_config["formatters"]["colourized"]["format"],
                style="{",
                use_colors=True,
            )
        )
    return handler



import logging
import sys
import time
from logging import Logger
from threading import Thread
from typing import Any, List, Tuple

import isar
from isar.apis.api import API
from isar.config.log import setup_loggers
from isar.config.open_telemetry import setup_open_telemetry
from isar.config.settings import robot_settings, settings
from isar.models.events import Events
from isar.modules import ApplicationContainer, get_injector
from isar.robot.robot import Robot
from isar.services.service_connections.mqtt.mqtt_client import MqttClient
from isar.services.service_connections.mqtt.robot_heartbeat_publisher import (
    RobotHeartbeatPublisher,
)
from isar.services.service_connections.mqtt.robot_info_publisher import (
    RobotInfoPublisher,
)
from isar.state_machine.state_machine import StateMachine, main
from isar.storage.uploader import Uploader
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.robot_interface import RobotInterface


def print_setting(
    setting: str = "", value: Any = "", fillchar: str = " ", width: int = 48
) -> None:
    separator = ": " if value != "" else ""
    text = setting.ljust(22, fillchar) + separator + str(value)
    print("*", text.ljust(width - 4, fillchar), "*")


def print_startup_info() -> None:
    logger: Logger = logging.getLogger("main")
    logger.info(
        """
         __   ________   ___        ________
        / /  / ______/  /   |      / ____  /
       / /  / /_____   / /| |     / /___/ /
      / /  /_____  /  / __  |    / __  __/
     / /  ______/ /  / /  | |   / /  | |
    /_/  /_______/  /_/   |_|  /_/   |_|

"""
    )

    def print_setting(setting: str = "", value: Any = "", fillchar: str = " ") -> None:
        separator = ": " if value != "" else ""
        logger.info(setting + separator + str(value))

    logger.info(
        f"Integration and Supervisory control of Autonomous Robots - Version: {isar.__version__}\n"
    )

    print_setting("ISAR settings")
    print_setting("ISAR ID", settings.ISAR_ID)
    print_setting("Robot package", settings.ROBOT_PACKAGE)
    print_setting("Robot name", settings.ROBOT_NAME)
    print_setting("Running on port", settings.API_PORT)
    print_setting("Using local storage", settings.STORAGE_LOCAL_ENABLED)
    print_setting("Using blob storage", settings.STORAGE_BLOB_ENABLED)
    print_setting("Blob storage account data", settings.BLOB_STORAGE_ACCOUNT_DATA)
    print_setting(
        "Blob storage account metadata", settings.BLOB_STORAGE_ACCOUNT_METADATA
    )
    print_setting("Using async inspection uploading", settings.UPLOAD_INSPECTIONS_ASYNC)
    print_setting("Plant code", settings.PLANT_CODE)
    print_setting("Plant name", settings.PLANT_NAME)
    print_setting("Plant shortname", settings.PLANT_SHORT_NAME)
    print_setting("Robot capabilities", robot_settings.CAPABILITIES)


def start() -> None:
    injector: ApplicationContainer = get_injector()

    setup_loggers()
    setup_open_telemetry(app=injector.api().app)
    logger: Logger = logging.getLogger("main")

    print_startup_info()

    state_machine: StateMachine = injector.state_machine()
    uploader: Uploader = injector.uploader()
    robot_interface: RobotInterface = injector.robot_interface()
    events: Events = injector.events()
    robot: Robot = injector.robot()

    threads: List[Thread] = []

    api: API = injector.api()
    api_thread: Thread = Thread(target=api.server.run, name="ISAR API", daemon=True)
    api_thread.start()
    threads.append(api_thread)

    api.wait_for_api_server_ready()

    state_machine_thread: Thread = Thread(
        target=main, name="ISAR State Machine", args=[state_machine], daemon=True
    )
    state_machine_thread.start()
    threads.append(state_machine_thread)

    uploader_thread: Thread = Thread(
        target=uploader.run, name="ISAR Uploader", daemon=True
    )
    uploader_thread.start()
    threads.append(uploader_thread)

    robot_service_thread: Thread = Thread(
        target=robot.run, name="Robot service", daemon=True
    )
    robot_service_thread.start()
    threads.append(robot_service_thread)

    if settings.UPLOAD_INSPECTIONS_ASYNC:

        def inspections_callback(inspection: Inspection, mission: Mission) -> None:
            message: Tuple[Inspection, Mission] = (
                inspection,
                mission,
            )
            state_machine.events.upload_queue.put(message)

        robot.register_and_monitor_inspection_callback(inspections_callback)

    mqtt_client: MqttClient = MqttClient(mqtt_queue=events.mqtt_queue)

    mqtt_thread: Thread = Thread(
        target=mqtt_client.run, name="ISAR MQTT Client", daemon=True
    )
    mqtt_thread.start()
    threads.append(mqtt_thread)

    robot_info_publisher: RobotInfoPublisher = RobotInfoPublisher(
        mqtt_queue=events.mqtt_queue
    )
    robot_info_thread: Thread = Thread(
        target=robot_info_publisher.run,
        name="ISAR Robot Info Publisher",
        daemon=True,
    )
    robot_info_thread.start()
    threads.append(robot_info_thread)

    robot_heartbeat_publisher: RobotHeartbeatPublisher = RobotHeartbeatPublisher(
        mqtt_queue=events.mqtt_queue
    )

    robot_heartbeat_thread: Thread = Thread(
        target=robot_heartbeat_publisher.run,
        name="ISAR Robot Heartbeat Publisher",
        daemon=True,
    )
    robot_heartbeat_thread.start()
    threads.append(robot_heartbeat_thread)

    publishers: List[Thread] = robot_interface.get_telemetry_publishers(
        queue=events.mqtt_queue,
        robot_name=settings.ROBOT_NAME,
        isar_id=settings.ISAR_ID,
    )
    for publisher in publishers:
        publisher.start()
        threads.append(publisher)

    while True:
        for thread in threads:
            if not thread.is_alive():
                logger.critical("Thread '%s' failed - ISAR shutting down", thread.name)
                sys.exit(1)
        time.sleep(settings.FSM_SLEEP_TIME)



import time
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field

from isar.apis.models.models import InputPose, InputPosition
from isar.config.settings import settings
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    RecordAudio,
    ReturnToHome,
    TakeCO2Measurement,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
    ZoomDescription,
)
from robot_interface.utilities.uuid_string_factory import uuid4_string


class InspectionTypes(str, Enum):
    image = "Image"
    thermal_image = "ThermalImage"
    video = "Video"
    thermal_video = "ThermalVideo"
    audio = "Audio"
    co2_measurement = "CO2Measurement"


class TaskType(str, Enum):
    Inspection = "inspection"
    ReturnToHome = "return_to_home"


class StartMissionInspectionDefinition(BaseModel):
    type: InspectionTypes = Field(default=InspectionTypes.image)
    inspection_target: InputPosition
    inspection_description: Optional[str] = None
    duration: Optional[float] = None


class StartMissionTaskDefinition(BaseModel):
    id: Optional[str] = None
    type: TaskType = Field(default=TaskType.Inspection)
    pose: InputPose
    inspection: Optional[StartMissionInspectionDefinition] = None
    tag: Optional[str] = None
    zoom: Optional[ZoomDescription] = None


class StartMissionDefinition(BaseModel):
    id: Optional[str] = None
    tasks: List[StartMissionTaskDefinition]
    name: Optional[str] = None
    start_pose: Optional[InputPose] = None


class StopMissionDefinition(BaseModel):
    mission_id: Optional[str] = None


class MissionFormatError(Exception):
    pass


def to_isar_mission(
    start_mission_definition: StartMissionDefinition,
) -> Mission:
    isar_tasks: List[TASKS] = []

    for task_definition in start_mission_definition.tasks:
        task: TASKS = to_isar_task(task_definition)
        isar_tasks.append(task)

    if not isar_tasks:
        raise MissionFormatError("Mission does not contain any valid tasks")

    isar_mission_name: str = (
        start_mission_definition.name
        if start_mission_definition.name
        else _build_mission_name()
    )

    start_pose = None
    if start_mission_definition.start_pose:
        start_pose = start_mission_definition.start_pose.to_alitra_pose()

    id = start_mission_definition.id if start_mission_definition.id else uuid4_string()

    return Mission(
        id=id,
        tasks=isar_tasks,
        name=isar_mission_name,
        start_pose=start_pose,
    )


def to_isar_task(task_definition: StartMissionTaskDefinition) -> TASKS:
    if task_definition.type == TaskType.Inspection:
        return to_inspection_task(task_definition)
    elif task_definition.type == TaskType.ReturnToHome:
        return ReturnToHome()
    else:
        raise MissionFormatError(
            f"Failed to create task: '{task_definition.type}' is not a valid"
        )


def to_inspection_task(task_definition: StartMissionTaskDefinition) -> TASKS:
    inspection_definition = task_definition.inspection

    if inspection_definition.type == InspectionTypes.image:
        return TakeImage(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
            target=task_definition.inspection.inspection_target.to_alitra_position(),
            zoom=task_definition.zoom,
        )
    elif inspection_definition.type == InspectionTypes.video:
        return TakeVideo(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
            target=task_definition.inspection.inspection_target.to_alitra_position(),
            duration=inspection_definition.duration,
            zoom=task_definition.zoom,
        )
    elif inspection_definition.type == InspectionTypes.thermal_image:
        return TakeThermalImage(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
            target=task_definition.inspection.inspection_target.to_alitra_position(),
            zoom=task_definition.zoom,
        )
    elif inspection_definition.type == InspectionTypes.thermal_video:
        return TakeThermalVideo(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
            target=task_definition.inspection.inspection_target.to_alitra_position(),
            duration=inspection_definition.duration,
            zoom=task_definition.zoom,
        )
    elif inspection_definition.type == InspectionTypes.audio:
        return RecordAudio(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
            target=task_definition.inspection.inspection_target.to_alitra_position(),
            duration=inspection_definition.duration,
        )
    elif inspection_definition.type == InspectionTypes.co2_measurement:
        return TakeCO2Measurement(
            id=task_definition.id if task_definition.id else uuid4_string(),
            robot_pose=task_definition.pose.to_alitra_pose(),
            tag_id=task_definition.tag,
            inspection_description=task_definition.inspection.inspection_description,
        )
    else:
        raise ValueError(
            f"Inspection type '{inspection_definition.type}' not supported"
        )


def _build_mission_name() -> str:
    return f"{settings.PLANT_SHORT_NAME}{settings.ROBOT_NAME}{int(time.time())}"



from typing import TYPE_CHECKING, List

import isar.state_machine.states.return_home_paused as ReturnHomePaused
import isar.state_machine.states.returning_home as ReturningHome
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class PausingReturnHome(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        def _failed_pause_event_handler(
            error_message: ErrorMessage,
        ) -> Transition[ReturningHome.ReturningHome]:
            state_machine.logger.warning(
                f"Failed to pause return home: {error_message.error_description}"
            )
            return ReturningHome.transition_to_existing_mission()

        def _successful_pause_event_handler(
            successful_pause: EmptyMessage,
        ) -> Transition[ReturnHomePaused.ReturnHomePaused]:
            return ReturnHomePaused.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="failed_pause_event",
                event=events.robot_service_events.mission_failed_to_pause,
                handler=_failed_pause_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="successful_pause_event",
                event=events.robot_service_events.mission_successfully_paused,
                handler=_successful_pause_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.PausingReturnHome,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[PausingReturnHome]:
    def _transition(state_machine: "StateMachine") -> PausingReturnHome:
        return PausingReturnHome(state_machine)

    return _transition



from collections import deque
from queue import Empty, Queue
from threading import Lock
from typing import Generic, Optional, Tuple, TypeVar

from isar.apis.models.models import (
    ControlMissionResponse,
    LockdownResponse,
    MaintenanceResponse,
    MissionStartResponse,
)
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus
from robot_interface.models.mission.task import TASKS

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")


class EmptyMessage:
    def __str__(self) -> str:
        return "Empty message"


class Event(Queue[T]):
    def __init__(self, name: str) -> None:
        super().__init__(maxsize=1)
        self.name = name

    def trigger_event(self, data: T, timeout: int = None) -> None:
        try:
            # We always want a timeout when blocking for results, so that
            # the thread will never get stuck waiting for a result
            self.put(data, block=timeout is not None, timeout=timeout)
        except Exception:
            if timeout is not None:
                raise EventTimeoutError
            return None

    def consume_event(self, timeout: int = None) -> Optional[T]:
        try:
            return self.get(block=timeout is not None, timeout=timeout)
        except Empty:
            if timeout is not None:
                raise EventTimeoutError
            return None
        except ValueError:
            raise EventConflictError

    def clear_event(self) -> None:
        while True:
            try:
                self.get(block=False)
            except Empty:
                break
            except ValueError:
                break

    def has_event(self) -> bool:
        return (
            self.qsize() != 0
        )  # Queue size is not reliable, but should be sufficient for this case

    def check(self) -> Optional[T]:
        if not self._qsize():
            return None
        with self.mutex:
            queueList = list(self.queue)
            return queueList.pop()

    def update(self, item: T) -> None:
        with self.mutex:
            self.queue: deque[T] = deque()
            self.queue.append(item)


class Events:
    def __init__(self) -> None:
        self.api_requests: APIRequests = APIRequests()
        self.state_machine_events: StateMachineEvents = StateMachineEvents()
        self.robot_service_events: RobotServiceEvents = RobotServiceEvents()

        self.upload_queue: Queue = Queue(maxsize=10)

        self.mqtt_queue: Queue = Queue()


class APIEvent(Generic[T1, T2]):
    """
    Creates request and response event. The events are defined such that the request is from
    api to state machine while the response is from state machine to api.
    """

    def __init__(self, name: str):
        self.request: Event[T1] = Event("api-" + name + "-request")
        self.response: Event[T2] = Event("api-" + name + "-request")
        self.lock: Lock = Lock()


class APIRequests:
    def __init__(self) -> None:
        self.start_mission: APIEvent[Mission, MissionStartResponse] = APIEvent(
            "start_mission"
        )
        self.stop_mission: APIEvent[str, ControlMissionResponse] = APIEvent(
            "stop_mission"
        )
        self.pause_mission: APIEvent[EmptyMessage, ControlMissionResponse] = APIEvent(
            "pause_mission"
        )
        self.resume_mission: APIEvent[EmptyMessage, ControlMissionResponse] = APIEvent(
            "resume_mission"
        )
        self.return_home: APIEvent[EmptyMessage, EmptyMessage] = APIEvent("return_home")
        self.release_intervention_needed: APIEvent[EmptyMessage, EmptyMessage] = (
            APIEvent("release_intervention_needed")
        )
        self.send_to_lockdown: APIEvent[EmptyMessage, LockdownResponse] = APIEvent(
            "send_to_lockdown"
        )
        self.release_from_lockdown: APIEvent[EmptyMessage, EmptyMessage] = APIEvent(
            "release_from_lockdown"
        )
        self.set_maintenance_mode: APIEvent[EmptyMessage, MaintenanceResponse] = (
            APIEvent("set_maintenance_mode")
        )
        self.release_from_maintenance_mode: APIEvent[EmptyMessage, EmptyMessage] = (
            APIEvent("release_from_maintenance_mode")
        )


class StateMachineEvents:
    def __init__(self) -> None:
        self.start_mission: Event[Mission] = Event("start_mission")
        self.stop_mission: Event[EmptyMessage] = Event("stop_mission")
        self.pause_mission: Event[EmptyMessage] = Event("pause_mission")
        self.resume_mission: Event[EmptyMessage] = Event("resume_mission")


class RobotServiceEvents:
    def __init__(self) -> None:
        self.mission_succeeded: Event[EmptyMessage] = Event("mission_succeeded")
        self.mission_started: Event[EmptyMessage] = Event("mission_started")
        self.mission_failed: Event[ErrorMessage] = Event("mission_failed")
        self.robot_status_changed: Event[EmptyMessage] = Event("robot_status_changed")
        self.mission_failed_to_stop: Event[ErrorMessage] = Event(
            "mission_failed_to_stop"
        )
        self.mission_successfully_stopped: Event[EmptyMessage] = Event(
            "mission_successfully_stopped"
        )
        self.mission_failed_to_pause: Event[ErrorMessage] = Event(
            "mission_failed_to_pause"
        )
        self.mission_successfully_paused: Event[EmptyMessage] = Event(
            "mission_successfully_paused"
        )
        self.mission_failed_to_resume: Event[ErrorMessage] = Event(
            "mission_failed_to_resume"
        )
        self.mission_successfully_resumed: Event[EmptyMessage] = Event(
            "mission_successfully_resumed"
        )
        self.request_inspection_upload: Event[Tuple[TASKS, Mission]] = Event(
            "request_inspection_upload"
        )
        self.robot_already_home: Event[EmptyMessage] = Event("robot_already_home")


class SharedState:
    def __init__(self) -> None:
        self.state: Event[States] = Event("state")
        self.robot_status: Event[RobotStatus] = Event("robot_status")
        self.robot_battery_level: Event[float] = Event("robot_battery_level")


class EventTimeoutError(Exception):
    pass


class EventConflictError(Exception):
    pass



from http import HTTPStatus
from threading import Thread

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.models.events import APIEvent, Event, EventTimeoutError
from isar.modules import ApplicationContainer
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states_enum import States
from tests.test_mocks.mission_definition import DummyMissionDefinition


def test_timeout_send_command(
    mocker: MockerFixture, scheduling_utilities: SchedulingUtilities
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    mocker.patch.object(Event, "consume_event", side_effect=EventTimeoutError)
    q: APIEvent = APIEvent("test")
    with pytest.raises(EventTimeoutError):
        scheduling_utilities._send_command(True, q)
    assert q.request.empty()


def test_robot_capable_of_mission(scheduling_utilities: SchedulingUtilities) -> None:
    assert scheduling_utilities.verify_robot_capable_of_mission(
        mission=DummyMissionDefinition.default_mission,
        robot_capabilities=["return_to_home", "take_image"],
    )


def test_robot_not_capable_of_mission(
    scheduling_utilities: SchedulingUtilities,
) -> None:
    with pytest.raises(HTTPException) as err:
        scheduling_utilities.verify_robot_capable_of_mission(
            mission=DummyMissionDefinition.default_mission,
            robot_capabilities=["return_to_home"],
        )
    assert err.value.status_code == HTTPStatus.BAD_REQUEST


def test_state_machine_ready_to_receive_mission(
    scheduling_utilities: SchedulingUtilities,
) -> None:
    assert scheduling_utilities.verify_state_machine_ready_to_receive_mission(
        States.Home
    )
    assert scheduling_utilities.verify_state_machine_ready_to_receive_mission(
        States.ReturningHome
    )
    assert scheduling_utilities.verify_state_machine_ready_to_receive_mission(
        States.AwaitNextMission
    )


def test_state_machine_not_ready_to_receive_mission(
    scheduling_utilities: SchedulingUtilities,
) -> None:
    with pytest.raises(HTTPException) as err:
        scheduling_utilities.verify_state_machine_ready_to_receive_mission(
            States.Monitor
        )
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_mission_already_started_causes_conflict(
    scheduling_utilities: SchedulingUtilities,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    start_mission_thread: Thread = Thread(
        target=scheduling_utilities.start_mission,
        args=[DummyMissionDefinition.default_mission],
    )
    start_mission_thread.start()

    with pytest.raises(HTTPException) as err:
        scheduling_utilities.start_mission(DummyMissionDefinition.default_mission)
    start_mission_thread.join()
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_pause_mission_twice_causes_conflict(
    scheduling_utilities: SchedulingUtilities,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    pause_mission_thread: Thread = Thread(target=scheduling_utilities.pause_mission)
    pause_mission_thread.start()

    with pytest.raises(HTTPException) as err:
        scheduling_utilities.pause_mission()
    pause_mission_thread.join()
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_resume_mission_twice_causes_conflict(
    scheduling_utilities: SchedulingUtilities,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    resume_mission_thread: Thread = Thread(target=scheduling_utilities.resume_mission)
    resume_mission_thread.start()

    with pytest.raises(HTTPException) as err:
        scheduling_utilities.resume_mission()
    resume_mission_thread.join()
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_stop_mission_twice_causes_conflict(
    scheduling_utilities: SchedulingUtilities,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    stop_mission_thread: Thread = Thread(target=scheduling_utilities.stop_mission)
    stop_mission_thread.start()

    with pytest.raises(HTTPException) as err:
        scheduling_utilities.stop_mission()
    stop_mission_thread.join()
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_return_home_twice_causes_conflict(
    scheduling_utilities: SchedulingUtilities,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "QUEUE_TIMEOUT", 2)
    return_home_thread: Thread = Thread(target=scheduling_utilities.return_home)
    return_home_thread.start()

    with pytest.raises(HTTPException) as err:
        scheduling_utilities.return_home()
    return_home_thread.join()
    assert err.value.status_code == HTTPStatus.CONFLICT


def test_api_with_unsuccessful_return_home_stop(
    mocker: MockerFixture,
    container: ApplicationContainer,
    sync_state_machine: StateMachine,
) -> None:
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    stopped_mission_response: ControlMissionResponse = ControlMissionResponse(
        success=False, failure_reason="ISAR failed to stop mission"
    )
    mocker.patch.object(
        Event,
        "consume_event",
        return_value=stopped_mission_response,
    )

    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.stop_mission()

    assert exception_details.value.status_code == HTTPStatus.SERVICE_UNAVAILABLE.value



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.stopping_go_to_recharge import StoppingGoToRecharge


def test_monitor_goes_to_return_home_when_battery_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Monitor(sync_state_machine, "mission_id")
    monitor_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        monitor_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(10.0)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is StoppingGoToRecharge



from datetime import datetime
from typing import Optional, Type

from alitra import Pose, Position
from pydantic import BaseModel, Field


class InspectionMetadata(BaseModel):
    start_time: datetime
    robot_pose: Pose
    target_position: Position
    file_type: str
    tag_id: Optional[str] = None
    inspection_description: Optional[str] = None


class ImageMetadata(InspectionMetadata):
    pass


class ThermalImageMetadata(InspectionMetadata):
    pass


class VideoMetadata(InspectionMetadata):
    duration: float


class ThermalVideoMetadata(InspectionMetadata):
    duration: float


class AudioMetadata(InspectionMetadata):
    duration: float


class GasMeasurementMetadata(InspectionMetadata):
    pass


class Inspection(BaseModel):
    metadata: InspectionMetadata
    id: str = Field(frozen=True)

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return InspectionMetadata


class InspectionValue(Inspection):
    value: float = Field(frozen=True)
    unit: str = Field(frozen=True)


class InspectionBlob(Inspection):
    data: Optional[bytes] = Field(default=None, frozen=True)


class Image(InspectionBlob):
    metadata: ImageMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ImageMetadata


class ThermalImage(InspectionBlob):
    metadata: ThermalImageMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ThermalImageMetadata


class Video(InspectionBlob):
    metadata: VideoMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return VideoMetadata


class ThermalVideo(InspectionBlob):
    metadata: ThermalVideoMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ThermalVideoMetadata


class Audio(InspectionBlob):
    metadata: AudioMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return AudioMetadata


class GasMeasurement(InspectionValue):
    metadata: GasMeasurementMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return GasMeasurementMetadata


class CO2Measurement(GasMeasurement):
    pass



from alitra import Frame, Orientation, Pose, Position

from robot_interface.models.mission.status import TaskStatus
from robot_interface.models.mission.task import TakeImage


class StubTask:
    @staticmethod
    def take_image(status: TaskStatus = TaskStatus.NotStarted) -> TakeImage:
        target_pose = Position(x=1, y=1, z=1, frame=Frame("robot"))
        robot_pose = Pose(
            position=Position(x=0, y=0, z=1, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        )
        return TakeImage(target=target_pose, robot_pose=robot_pose, status=status)



# import warnings
# # This removes the deprecation warning in testcontainers that is not our fault
# warnings.filterwarnings("ignore", category=DeprecationWarning, module="testcontainers")

import logging
import shutil
from pathlib import Path
from typing import Generator

import pytest
import sqlalchemy
from dependency_injector.providers import Singleton
from dependency_injector.wiring import providers
from fastapi import FastAPI
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture
from testcontainers.mysql import MySqlContainer

from isar.apis.security.authentication import Authenticator
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import State
from isar.models.events import Events
from isar.modules import ApplicationContainer
from isar.robot.robot import Robot
from isar.robot.robot_battery import RobotBatteryThread
from isar.robot.robot_monitor_mission import RobotMonitorMissionThread
from isar.robot.robot_pause_mission import RobotPauseMissionThread
from isar.robot.robot_resume_mission import RobotResumeMissionThread
from isar.robot.robot_start_mission import RobotStartMissionThread
from isar.robot.robot_status import RobotStatusThread
from isar.robot.robot_stop_mission import RobotStopMissionThread
from isar.robot.robot_upload_inspection import RobotUploadInspectionThread
from isar.services.service_connections.persistent_memory import Base
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.storage.uploader import Uploader
from robot_interface.robot_interface import RobotInterface
from tests.test_mocks.blob_storage import StorageFake
from tests.test_mocks.robot_interface import StubRobot
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
    UploaderThreadMock,
)


@pytest.fixture(autouse=True)
def setup_test_environment() -> None:
    settings.PERSISTENT_STORAGE_CONNECTION_STRING = ""

    logger_names = [
        "main",
        "api",
        "state_machine",
        "robot",
        "mqtt_client",
        "uploader",
        "fastapi_azure_auth",
    ]
    loggers = [logging.getLogger(name) for name in logger_names]
    for logger in loggers:
        logger.propagate = False  # Turn off logging in the command line for testing


@pytest.fixture()
def container() -> ApplicationContainer:
    """Fixture to provide the dependency-injector container without auth."""
    container = ApplicationContainer()
    container.events.override(providers.Singleton(Events))
    container.storage_handlers.override(
        providers.List(providers.Singleton(StorageFake))
    )
    container.robot_interface.override(providers.Singleton(StubRobot))
    container.uploader.override(
        providers.Singleton(
            Uploader,
            container.events(),
            container.storage_handlers(),
            container.mqtt_client(),
        )
    )
    container.robot.override(
        providers.Singleton(
            Robot,
            events=container.events(),
            robot=container.robot_interface(),
            shared_state=container.shared_state(),
        )
    )
    return container


@pytest.fixture()
def app(container: ApplicationContainer) -> FastAPI:
    """Fixture to provide the FastAPI app."""
    container.authenticator.override(
        providers.Singleton(Authenticator, authentication_enabled=False)
    )

    app = container.api().get_app()
    return app


@pytest.fixture()
def client(app: FastAPI) -> TestClient:
    """Fixture to provide a test client for the FastAPI app."""
    client = TestClient(app)
    return client


@pytest.fixture()
def client_auth(container: ApplicationContainer) -> TestClient:
    """Fixture to provide a test client for the FastAPI app with auth."""
    container.authenticator.override(
        providers.Singleton(Authenticator, authentication_enabled=True)
    )

    app = container.api().get_app()
    client = TestClient(app)
    return client


@pytest.fixture()
def access_token() -> str:
    """Fixture to provide a dummy access token."""
    return "DummyToken"


@pytest.fixture()
def state_machine(
    container: ApplicationContainer, robot: RobotInterface
) -> StateMachine:
    """Fixture to provide the StateMachine instance."""
    return StateMachine(
        events=container.events(),
        shared_state=container.shared_state(),
        robot=robot,
        mqtt_publisher=container.mqtt_client(),
    )


@pytest.fixture()
def sync_state_machine(
    container: ApplicationContainer, robot: RobotInterface, mocker: MockerFixture
) -> StateMachine:
    """Fixture to provide the StateMachine instance without running the state loops."""
    mocker.patch.object(State, "run", return_value=lambda: None)
    return StateMachine(
        events=container.events(),
        shared_state=container.shared_state(),
        robot=robot,
        mqtt_publisher=container.mqtt_client(),
    )


@pytest.fixture()
def robot() -> StubRobot:
    """Fixture to provide a mock robot instance."""
    return StubRobot()


@pytest.fixture()
def scheduling_utilities(
    container: ApplicationContainer,
) -> Singleton[SchedulingUtilities]:
    """Fixture to provide the SchedulingUtilities instance."""
    return container.scheduling_utilities()


@pytest.fixture
def state_machine_thread(
    container: ApplicationContainer,
) -> Generator[StateMachineThreadMock, None, None]:
    state_machine_thread: StateMachineThreadMock = StateMachineThreadMock(
        container=container,
    )
    yield state_machine_thread
    state_machine_thread.join()


@pytest.fixture
def uploader_thread(
    container: ApplicationContainer,
) -> Generator[UploaderThreadMock, None, None]:
    uploader_thread: UploaderThreadMock = UploaderThreadMock(container=container)
    yield uploader_thread
    uploader_thread.join()


@pytest.fixture
def robot_service_thread(
    container: ApplicationContainer,
) -> Generator[RobotServiceThreadMock, None, None]:
    robot_service: Robot = Robot(
        events=container.events(),
        robot=container.robot_interface(),
        shared_state=container.shared_state(),
        mqtt_publisher=container.mqtt_client(),
    )

    robot_service_thread: RobotServiceThreadMock = RobotServiceThreadMock(
        robot_service=robot_service
    )
    yield robot_service_thread
    robot_service_thread.join()


@pytest.fixture
def mocked_robot_service(
    container: ApplicationContainer, mocker: MockerFixture
) -> Robot:
    robot_service: Robot = Robot(
        events=container.events(),
        robot=container.robot_interface(),
        shared_state=container.shared_state(),
        mqtt_publisher=container.mqtt_client(),
    )

    mocker.patch.object(RobotStartMissionThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotBatteryThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotStatusThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotMonitorMissionThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotStopMissionThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotPauseMissionThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotResumeMissionThread, "run", return_value=lambda: None)
    mocker.patch.object(RobotUploadInspectionThread, "run", return_value=lambda: None)

    mocker.patch.object(RobotStartMissionThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotBatteryThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotStatusThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotMonitorMissionThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotStopMissionThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotPauseMissionThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotResumeMissionThread, "join", return_value=lambda: None)
    mocker.patch.object(RobotUploadInspectionThread, "join", return_value=lambda: None)

    return robot_service


@pytest.fixture(autouse=True)
def run_before_and_after_tests() -> None:  # type: ignore
    results_folder: Path = Path("tests/results")
    yield

    print("Removing temporary results folder for testing")
    if results_folder.exists():
        shutil.rmtree(results_folder)
    print("Cleanup finished")


@pytest.fixture()
def setup_db_connection_string() -> Generator[None, None, None]:
    with MySqlContainer("mysql:9.4.0", dialect="pymysql") as mysql:
        connection_url = mysql.get_connection_url()
        settings.PERSISTENT_STORAGE_CONNECTION_STRING = connection_url

        engine = sqlalchemy.create_engine(connection_url)
        Base.metadata.create_all(engine)

        yield



from typing import Any, Optional


class StubRequests:
    def __init__(self, json_data: Optional[Any] = None) -> None:
        self.json_data = json_data

    def json(self) -> Optional[Any]:
        return self.json_data



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from pydantic import BaseModel

from robot_interface.models.inspection.inspection import InspectionBlob
from robot_interface.models.mission.mission import Mission


class BlobStoragePath(BaseModel):
    storage_account: str
    blob_container: str
    blob_name: str


class LocalStoragePath(BaseModel):
    file_path: Path


TPath = TypeVar("TPath", BlobStoragePath, LocalStoragePath)


class StoragePaths(BaseModel, Generic[TPath]):
    data_path: TPath
    metadata_path: TPath


class StorageInterface(metaclass=ABCMeta):
    @abstractmethod
    def store(self, inspection: InspectionBlob, mission: Mission) -> StoragePaths:
        """
        Parameters
        ----------
        inspection : InspectionBlob
            The inspection object to be stored.
        mission : Mission
            Mission the inspection is a part of.

        Returns
        ----------
        StoragePaths
            Paths to the data and metadata

        Raises
        ----------
        StorageException
            An error occurred when storing the inspection.
        """
        pass


class StorageException(Exception):
    pass



from typing import TYPE_CHECKING, List, Union

import isar.state_machine.states.home as Home
import isar.state_machine.states.intervention_needed as InterventionNeeded
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Maintenance(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        def _release_from_maintenance_handler(
            should_release_from_maintenance: EmptyMessage,
        ) -> Union[
            Transition[Home.Home], Transition[InterventionNeeded.InterventionNeeded]
        ]:
            events.api_requests.release_from_maintenance_mode.response.trigger_event(
                EmptyMessage()
            )

            robot_status = state_machine.shared_state.robot_status.check()
            if robot_status == RobotStatus.Home:
                return Home.transition()
            else:
                return InterventionNeeded.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[EmptyMessage](
                name="release_from_maintenance",
                event=events.api_requests.release_from_maintenance_mode.request,
                handler=_release_from_maintenance_handler,
            ),
        ]

        super().__init__(
            state_name=States.Maintenance,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[Maintenance]:
    def _transition(state_machine: "StateMachine") -> Maintenance:
        return Maintenance(state_machine)

    return _transition



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.return_home_paused import ReturnHomePaused
from isar.state_machine.states.stopping_paused_return_home import (
    StoppingPausedReturnHome,
)
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from tests.test_mocks.task import StubTask


def test_stopping_paused_return_home_mission_fails(
    sync_state_machine: StateMachine,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    sync_state_machine.current_state = StoppingPausedReturnHome(
        sync_state_machine, mission
    )
    stopping_paused_return_home_state: State = cast(
        State, sync_state_machine.current_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        stopping_paused_return_home_state.get_event_handler_by_name("failed_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(error_description="", error_reason=ErrorReason.RobotAPIException)
    )

    assert sync_state_machine.events.api_requests.start_mission.response.has_event()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturnHomePaused


def test_stopping_paused_return_home_mission_succeeds(
    sync_state_machine: StateMachine,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    sync_state_machine.shared_state.robot_battery_level.trigger_event(90.0)
    sync_state_machine.current_state = StoppingPausedReturnHome(
        sync_state_machine, mission
    )
    stopping_paused_return_home_state: State = cast(
        State, sync_state_machine.current_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        stopping_paused_return_home_state.get_event_handler_by_name(
            "successful_stop_event"
        )
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor



from typing import TYPE_CHECKING, List, Union

import isar.state_machine.states.home as Home
import isar.state_machine.states.recharging as Recharging
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Lockdown(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        def _release_from_lockdown_handler(
            should_release_from_lockdown: EmptyMessage,
        ) -> Union[Transition[Home.Home], Transition[Recharging.Recharging]]:
            events.api_requests.release_from_lockdown.response.trigger_event(
                EmptyMessage()
            )
            if state_machine.battery_level_is_above_mission_start_threshold():
                return Home.transition()
            else:
                return Recharging.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[EmptyMessage](
                name="release_from_lockdown",
                event=events.api_requests.release_from_lockdown.request,
                handler=_release_from_lockdown_handler,
            ),
        ]

        super().__init__(
            state_name=States.Lockdown,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[Lockdown]:
    def _transition(state_machine: "StateMachine") -> Lockdown:
        return Lockdown(state_machine)

    return _transition



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.home as Home
import isar.state_machine.states.lockdown as Lockdown
import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.offline as Offline
from isar.apis.models.models import LockdownResponse, MaintenanceResponse
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Recharging(State):

    def __init__(self, state_machine: "StateMachine"):
        shared_state = state_machine.shared_state
        events = state_machine.events

        def robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[Home.Home]]:
            if battery_level < settings.ROBOT_BATTERY_RECHARGE_THRESHOLD:
                return None

            return Home.transition()

        def robot_offline_handler(
            robot_status: RobotStatus,
        ) -> Optional[Transition[Offline.Offline]]:
            if robot_status == RobotStatus.Offline:
                self.logger.info(
                    "Got robot status offline while in recharging state. Leaving recharging state."
                )
                return Offline.transition()
            return None

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[Lockdown.Lockdown]:
            events.api_requests.send_to_lockdown.response.trigger_event(
                LockdownResponse(lockdown_started=True)
            )
            return Lockdown.transition()

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[Maintenance.Maintenance]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[RobotStatus](
                name="robot_offline_event",
                event=shared_state.robot_status,
                handler=robot_offline_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Recharging,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[Recharging]:
    def _transition(state_machine: "StateMachine") -> Recharging:
        return Recharging(state_machine)

    return _transition



import time
from collections import deque
from http import HTTPStatus
from typing import Optional, cast
from uuid import uuid4

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.modules import ApplicationContainer
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.pausing import Pausing
from isar.state_machine.states.resuming import Resuming
from isar.state_machine.states.stopping_go_to_lockdown import StoppingGoToLockdown
from isar.state_machine.states.stopping_go_to_recharge import StoppingGoToRecharge
from isar.state_machine.states.stopping_return_home import StoppingReturnHome
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    ErrorReason,
    RobotException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import ReturnToHome
from tests.test_mocks.robot_interface import (
    StubRobot,
    StubRobotMissionStatusRaisesException,
)
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
)
from tests.test_mocks.task import StubTask


def _mock_robot_exception_with_message() -> RobotException:
    raise RobotException(
        error_reason=ErrorReason.RobotUnknownErrorException,
        error_description="This is an example error description",
    )


def test_stopping_to_recharge_goes_to_monitor(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = StoppingGoToRecharge(
        sync_state_machine, "mission_id"
    )
    stopping_go_to_recharge_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_go_to_recharge_state.get_event_handler_by_name("failed_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.mqtt_queue.empty()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor


def test_transitioning_to_monitor_from_stopping_when_return_home_cancelled(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    example_mission: Mission = Mission(
        name="Dummy return home misson", tasks=[ReturnToHome()]
    )
    sync_state_machine.current_state = StoppingReturnHome(
        sync_state_machine, example_mission
    )

    stopping_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())
    sync_state_machine.current_state = transition(sync_state_machine)

    assert type(sync_state_machine.current_state) is Monitor


def test_stopping_lockdown_failing_to_monitor(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = StoppingGoToLockdown(
        sync_state_machine, "mission_id"
    )

    stopping_go_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_go_to_lockdown_state.get_event_handler_by_name("failed_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert (
        not sync_state_machine.events.api_requests.send_to_lockdown.response.check().lockdown_started
    )

    assert sync_state_machine.events.mqtt_queue.empty()

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor


def test_transition_from_pausing_to_monitor(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Pausing(sync_state_machine, "mission_id")

    pausing_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        pausing_state.get_event_handler_by_name("failed_pause_event")
    )

    assert event_handler is not None

    error_event = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description=""
    )
    transition = event_handler.handler(error_event)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor


def test_transition_from_resuming_to_monitor(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Resuming(sync_state_machine, "mission_id")

    resuming_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        resuming_state.get_event_handler_by_name("successful_resume_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Monitor


def test_state_machine_with_unsuccessful_mission_stop(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mocker.patch.object(settings, "FSM_SLEEP_TIME", 0.01)
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(
        StubRobot, "mission_status", return_value=MissionStatus.InProgress
    )
    mocker.patch.object(
        StubRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(0.5)
    scheduling_utilities.stop_mission()
    time.sleep(2)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
            States.Stopping,
            States.Monitor,
        ]
    )


def test_state_machine_with_unsuccessful_mission_stop_with_mission_id(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(
        StubRobot, "mission_status", return_value=MissionStatus.InProgress
    )
    mocker.patch.object(
        StubRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    settings.FSM_SLEEP_TIME = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.start_mission(mission=mission)
    time.sleep(1)
    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.stop_mission(str(uuid4()))

    assert exception_details.value.status_code == HTTPStatus.SERVICE_UNAVAILABLE.value
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
        ]
    )


def test_robot_mission_status_exception_handling(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:

    mission = Mission(
        name="Dummy mission",
        tasks=[StubTask.take_image()],
    )
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    robot_service_thread.robot_service.robot = StubRobotMissionStatusRaisesException()

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.start_mission(mission=mission)

    time.sleep(1)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
            States.AwaitNextMission,
        ]
    )



import logging
from pathlib import Path

from azure.core.exceptions import ResourceExistsError
from azure.storage.blob import BlobServiceClient, ContainerClient

from isar.config.settings import settings
from isar.storage.storage_interface import (
    BlobStoragePath,
    StorageException,
    StorageInterface,
    StoragePaths,
)
from isar.storage.utilities import construct_metadata_file, construct_paths
from robot_interface.models.inspection.inspection import InspectionBlob
from robot_interface.models.mission.mission import Mission


class BlobStorage(StorageInterface):
    def __init__(self) -> None:
        self.logger = logging.getLogger("uploader")

        self.container_client_data = self._get_container_client(
            settings.BLOB_STORAGE_ACCOUNT_DATA,
            settings.BLOB_STORAGE_CONNECTION_STRING_DATA,
        )
        self.container_client_metadata = self._get_container_client(
            settings.BLOB_STORAGE_ACCOUNT_METADATA,
            settings.BLOB_STORAGE_CONNECTION_STRING_METADATA,
        )

    def _get_container_client(
        self, account_name: str, storage_connection_string: str
    ) -> ContainerClient:

        if not storage_connection_string:
            raise RuntimeError("BlobStorage connection string is Empty or None")

        try:
            blob_service_client = BlobServiceClient.from_connection_string(
                storage_connection_string
            )
            if blob_service_client.account_name != account_name:
                raise Exception(
                    f"The blob storage in the connection string does not match the settings for blob storage account {account_name}"
                )
        except Exception as e:
            self.logger.error("Unable to retrieve blob service client. Error: %s", e)
            raise e

        container_client: ContainerClient = blob_service_client.get_container_client(
            settings.BLOB_CONTAINER
        )

        if not container_client.exists():
            raise RuntimeError(
                "The configured blob container %s does not exist",
                settings.BLOB_CONTAINER,
            )
        return container_client

    def store(
        self, inspection: InspectionBlob, mission: Mission
    ) -> StoragePaths[BlobStoragePath]:
        if inspection.data is None:
            raise StorageException("Nothing to store. The inspection data is empty")

        data_filename, metadata_filename = construct_paths(
            inspection=inspection, mission=mission
        )

        metadata_bytes: bytes = construct_metadata_file(
            inspection=inspection, mission=mission, filename=data_filename.name
        )

        data_path = self._upload_file(
            filename=data_filename,
            data=inspection.data,
            container_client=self.container_client_data,
            account_name=settings.BLOB_STORAGE_ACCOUNT_DATA,
        )
        metadata_path = self._upload_file(
            filename=metadata_filename,
            data=metadata_bytes,
            container_client=self.container_client_metadata,
            account_name=settings.BLOB_STORAGE_ACCOUNT_METADATA,
        )
        return StoragePaths(data_path=data_path, metadata_path=metadata_path)

    def _upload_file(
        self,
        filename: Path,
        data: bytes,
        container_client: ContainerClient,
        account_name: str,
    ) -> BlobStoragePath:
        blob_client = container_client.get_blob_client(filename.as_posix())
        try:
            blob_client.upload_blob(data=data)
        except ResourceExistsError as e:
            self.logger.error(
                "Blob %s already exists in container. Error: %s", filename.as_posix(), e
            )
            raise StorageException from e
        except Exception as e:
            self.logger.error(
                "An unexpected error occurred while uploading blob: %s", e
            )
            raise StorageException from e

        return BlobStoragePath(
            storage_account=account_name,
            blob_container=settings.BLOB_CONTAINER,
            blob_name=blob_client.blob_name,
        )



import logging
import time
from threading import Event, Thread
from typing import Optional

from isar.config.settings import settings
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    ErrorReason,
    RobotActionException,
    RobotException,
    RobotNoMissionRunningException,
)
from robot_interface.robot_interface import RobotInterface


class RobotPauseMissionThread(Thread):
    def __init__(
        self,
        robot: RobotInterface,
        signal_thread_quitting: Event,
    ):
        self.logger = logging.getLogger("robot")
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.error_message: Optional[ErrorMessage] = None
        Thread.__init__(self, name="Robot pause mission thread")

    def run(self) -> None:
        retries = 0
        error: Optional[ErrorMessage] = None
        while retries < settings.STATE_TRANSITION_NUM_RETIRES:
            if self.signal_thread_quitting.wait(0):
                return
            try:
                self.robot.pause()
            except RobotNoMissionRunningException as e:
                error = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                break
            except (RobotActionException, RobotException) as e:
                self.logger.warning(
                    f"\nFailed to pause robot because: {e.error_description}"
                    f"\nAttempting to pause the robot again"
                )
                retries += 1
                error = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                time.sleep(settings.FSM_SLEEP_TIME)
                continue
            except Exception as e:
                self.logger.error(
                    f"\nAn unexpected error occurred while pausing the robot: {str(e)}"
                )
                error = ErrorMessage(
                    error_reason=ErrorReason.RobotUnknownErrorException,
                    error_description=str(e),
                )
                break
            return

        error_description = (
            f"\nFailed to pause the robot after {retries} attempts because: "
            f"{error.error_description}"
            f"\nBe aware that the robot may still be moving even though a pause has "
            "been attempted"
        )

        self.error_message = ErrorMessage(
            error_reason=error.error_reason,
            error_description=error_description,
        )



from http import HTTPStatus

import jwt
import pytest
from fastapi.testclient import TestClient


def stub_access_token() -> str:
    token = jwt.encode(payload={}, key="some_key")

    return token


class TestAuthentication:
    @pytest.mark.parametrize(
        "query_string",
        ["start-mission?ID=1", "stop-mission"],
    )
    def test_authentication(
        self,
        client_auth: TestClient,
        query_string: str,
    ) -> None:
        token = stub_access_token()
        expected_status_code = HTTPStatus.UNAUTHORIZED

        response = client_auth.post(
            f"schedule/{query_string}",
            headers={"Authorization": "Bearer " + token},
        )

        assert response.status_code == expected_status_code



from enum import Enum


class States(str, Enum):
    Monitor = "monitor"
    ReturningHome = "returning_home"
    Stopping = "stopping"
    StoppingReturnHome = "stopping_return_home"
    Paused = "paused"
    Pausing = "pausing"
    Resuming = "resuming"
    PausingReturnHome = "pausing_return_home"
    ResumingReturnHome = "resuming_return_home"
    ReturnHomePaused = "return_home_paused"
    AwaitNextMission = "await_next_mission"
    Home = "home"
    Offline = "offline"
    BlockedProtectiveStop = "blocked_protective_stop"
    UnknownStatus = "unknown_status"
    InterventionNeeded = "intervention_needed"
    Recharging = "recharging"
    StoppingGoToLockdown = "stopping_go_to_lockdown"
    GoingToLockdown = "going_to_lockdown"
    Lockdown = "lockdown"
    GoingToRecharging = "going_to_recharging"
    StoppingGoToRecharge = "stopping_go_to_recharge"
    Maintenance = "maintenance"
    StoppingDueToMaintenance = "stopping_due_to_maintenance"
    StoppingPausedMission = "stopping_paused_mission"
    StoppingPausedReturnHome = "stopping_paused_return_home"

    def __repr__(self) -> str:
        return self.value



from typing import Optional, cast

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.going_to_lockdown import GoingToLockdown
from isar.state_machine.states.going_to_recharging import GoingToRecharging
from isar.state_machine.states.intervention_needed import InterventionNeeded
from isar.state_machine.states.returning_home import ReturningHome
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.status import RobotStatus


def test_going_to_recharging_goes_to_intervention_needed(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = GoingToRecharging(sync_state_machine)
    going_to_recharging_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_recharging_state.get_event_handler_by_name("mission_failed_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException,
            error_description="test",
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is InterventionNeeded


def test_going_to_lockdown_task_failed_transitions_to_intervention_needed(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = GoingToLockdown(sync_state_machine)

    going_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_lockdown_state.get_event_handler_by_name("mission_failed_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException,
            error_description="test",
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is InterventionNeeded


def test_going_to_lockdown_mission_failed_transitions_to_intervention_needed(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = GoingToLockdown(sync_state_machine)

    going_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_lockdown_state.get_event_handler_by_name("mission_failed_event")
    )

    assert event_handler is not None

    # The type of error reason is not important for this test
    transition = event_handler.handler(
        ErrorMessage(error_description="", error_reason=ErrorReason.RobotAPIException)
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is InterventionNeeded


def test_state_machine_with_return_home_failure(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("mission_failed_event")
    )

    # We do not retry return home missions if the robot is not ready for another mission
    sync_state_machine.shared_state.robot_status.trigger_event(RobotStatus.Available)

    assert event_handler is not None

    for i in range(settings.RETURN_HOME_RETRY_LIMIT - 1):

        transition = event_handler.handler(
            ErrorMessage(
                error_reason=ErrorReason.RobotUnknownErrorException,
                error_description="test",
            )
        )

        assert transition is None  # type: ignore
        assert sync_state_machine.current_state.failed_return_home_attempts == i + 1

    transition = event_handler.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException,
            error_description="test",
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is InterventionNeeded


def test_intervention_needed_transitions_does_not_transition_if_status_is_not_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = InterventionNeeded(sync_state_machine)

    intervention_needed_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        intervention_needed_state.get_event_handler_by_name("robot_status_event")
    )
    assert event_handler is not None

    statuses = [
        RobotStatus.Available,
        RobotStatus.BlockedProtectiveStop,
        RobotStatus.Busy,
        RobotStatus.Paused,
        RobotStatus.Offline,
    ]
    for status in statuses:
        sync_state_machine.shared_state.robot_status.update(status)

        transition = event_handler.handler(EmptyMessage())

        assert transition is None  # type: ignore



# This file uses SQLAlchemy to interface the persistent database storage.

import sqlalchemy
from sqlalchemy import Boolean, Integer, String
from sqlalchemy.orm import DeclarativeBase, Mapped, Session, mapped_column


class Base(DeclarativeBase):
    pass


class PersistentRobotState(Base):
    __tablename__ = "persistent_robot_state"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    robot_id: Mapped[str] = mapped_column(String(64))
    is_maintenance_mode: Mapped[bool] = mapped_column(Boolean, nullable=False)

    def __repr__(self) -> str:
        return f"PersistentRobotState(id={self.id!r}, robot_id={self.robot_id!r}, is_maintenance_mode={self.is_maintenance_mode!r})"


class NoSuchRobotException(Exception):
    pass


def read_persistent_robot_state_is_maintenance_mode(
    connection_string: str, robot_id: str
) -> bool:
    engine = sqlalchemy.create_engine(connection_string)

    with Session(engine) as session:
        statement = sqlalchemy.select(PersistentRobotState).where(
            PersistentRobotState.robot_id == robot_id
        )
        read_persistent_state = session.scalar(statement)

        if read_persistent_state is None:
            raise NoSuchRobotException(
                f"No robot in persistent storage with id {robot_id}"
            )

        return read_persistent_state.is_maintenance_mode


def change_persistent_robot_state_is_maintenance_mode(
    connection_string: str, robot_id: str, value: bool
) -> None:
    engine = sqlalchemy.create_engine(connection_string)

    with Session(engine) as session:
        statement = sqlalchemy.select(PersistentRobotState).where(
            PersistentRobotState.robot_id == robot_id
        )
        read_persistent_state = session.scalar(statement)

        read_persistent_state.is_maintenance_mode = value
        session.commit()


def create_persistent_robot_state(connection_string: str, robot_id: str) -> None:
    engine = sqlalchemy.create_engine(connection_string)

    with Session(engine) as session:
        persistent_state = PersistentRobotState(
            robot_id=robot_id, is_maintenance_mode=True
        )
        session.add_all([persistent_state])
        session.commit()



import os
from importlib.resources import as_file, files
from typing import Any, List

from dotenv import load_dotenv
from pydantic import Field, ValidationInfo, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from robot_interface.models.robots.robot_model import RobotModel
from robot_interface.telemetry.payloads import DocumentInfo


class Settings(BaseSettings):
    # Endpoint open telemetry will export telemetry in the otlp protocol to
    OPEN_TELEMETRY_OTLP_EXPORTER_ENDPOINT: str = Field(default="http://localhost:4318")

    # Connection string for Azure Application Insights
    # This is optional and it will use managed identity if not set
    APPLICATIONINSIGHTS_CONNECTION_STRING: str = Field(default="")

    # Determines which robot package ISAR will attempt to import
    # Name must match with an installed python package in the local environment
    ROBOT_PACKAGE: str = Field(default="isar_robot")

    # Determines the local path in which results from missions are stored
    LOCAL_STORAGE_PATH: str = Field(default="./results")

    # Timeout in seconds for direct HTTP requests made through the RequestHandler
    REQUEST_TIMEOUT: int = Field(default=30)

    # Timeout in seconds for checking whether there is a message on a queue
    QUEUE_TIMEOUT: int = Field(default=10)

    # Sleep time for while loops in the finite state machine in seconds
    # The sleep is used to throttle the system on every iteration in the loop
    FSM_SLEEP_TIME: float = Field(default=0.1)

    # Name of default map transformation
    DEFAULT_MAP: str = Field(default="default_map")

    # Determines the number of state transitions that are kept in the log
    STATE_TRANSITIONS_LOG_LENGTH: int = Field(default=20)

    # Number of attempts to request a task status in monitor before cancelling
    REQUEST_STATUS_FAILURE_COUNTER_LIMIT: int = Field(default=3)

    # Time allocated to reconnect when failing to retrieve status due to communication
    # issues
    REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY: float = Field(default=10)

    # Number of attempts for state transitions resume and pause if failed
    STATE_TRANSITION_NUM_RETIRES: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    STOP_ROBOT_ATTEMPTS_LIMIT: int = Field(default=3)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_ATTEMPTS_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_MAX_WAIT: int = Field(default=60)

    # ISAR telemetry intervals
    ROBOT_HEARTBEAT_PUBLISH_INTERVAL: float = Field(default=1)
    ROBOT_INFO_PUBLISH_INTERVAL: float = Field(default=5)
    ROBOT_API_BATTERY_POLL_INTERVAL: float = Field(default=5)
    ROBOT_API_STATUS_POLL_INTERVAL: float = Field(default=5)
    THREAD_CHECK_INTERVAL: float = Field(default=0.01)

    # Determines the minimum battery level the robot must have to start a mission
    # If it drops below this level it will recharge to the value set by
    # ROBOT_BATTERY_RECHARGE_THRESHOLD before starting new missions
    ROBOT_MISSION_BATTERY_START_THRESHOLD: float = Field(default=25.0)

    # Determines the minimum battery threshold to consider the robot recharged
    # and ready for more missions, after having run low on charge
    ROBOT_BATTERY_RECHARGE_THRESHOLD: float = Field(default=80.0)

    # FastAPI host
    API_HOST_VIEWED_EXTERNALLY: str = Field(default="0.0.0.0")

    # FastAPI port
    API_PORT: int = Field(default=3000)

    # Determines how long delay time should be allowed before returning home
    RETURN_HOME_DELAY: int = Field(default=10)

    # Sets how many times the robot should try to return home if a return home fails
    RETURN_HOME_RETRY_LIMIT: int = Field(default=5)

    # Determines which storage modules are used by ISAR
    # Multiple storage modules can be chosen
    # Each module will be called when storing results from inspections
    # Selecting a different storage module than local may require certain access rights
    STORAGE_LOCAL_ENABLED: bool = Field(default=True)
    STORAGE_BLOB_ENABLED: bool = Field(default=False)

    # Determines whether certificate based encryption will be used for the MQTT
    # communication.
    MQTT_SSL_ENABLED: bool = Field(default=True)

    # Determines whether authentication is enabled for the API or not
    # Enabling this requires certain resources available for OAuth2 authentication
    # Currently supported authentication is Azure AD
    # (https://github.com/Intility/fastapi-azure-auth)
    AUTHENTICATION_ENABLED: bool = Field(default=True)

    # Tenant ID for the Azure tenant with your Azure Active Directory application
    AZURE_TENANT_ID: str = Field(default="3aa4a235-b6e2-48d5-9195-7fcf05b459b0")

    # Client ID for ISAR
    AZURE_CLIENT_ID: str = Field(default="fd384acd-5c1b-4c44-a1ac-d41d720ed0fe")
    # If AZURE_CLIENT_ID is set as an environment variable, overwrite this despite missing prefix.
    # This is done to avoid double config of ISAR_AZURE_CLIENT_ID and AZURE_CLIENT_ID.
    # We need the latter as an environment variable for the EnvironmentCredential method for AzureAD.
    azure_client_id_name: str = "AZURE_CLIENT_ID"
    if os.environ.get(azure_client_id_name) is not None:
        print("Using environment variable for AZURE_CLIENT_ID")
        AZURE_CLIENT_ID = os.environ[azure_client_id_name]

    # MQTT username
    # The username and password is set by the MQTT broker and must be known in advance
    # The password should be set as an environment variable "MQTT_PASSWORD"
    # If the password is not set in the environment an empty string will be used
    MQTT_USERNAME: str = Field(default="isar")
    MQTT_HOST: str = Field(default="localhost")
    MQTT_PORT: int = Field(default=1883)

    # Determines whether inspections are uploaded asynchronously or get_inspections in robotinterface
    UPLOAD_INSPECTIONS_ASYNC: bool = Field(default=False)

    # URL to storage account for Azure Blob Storage
    BLOB_STORAGE_ACCOUNT_DATA: str = Field(default="")
    BLOB_STORAGE_ACCOUNT_METADATA: str = Field(default="")

    # Blob storage connection strings for Azure Blob Storage
    BLOB_STORAGE_CONNECTION_STRING_DATA: str = Field(default="")
    BLOB_STORAGE_CONNECTION_STRING_METADATA: str = Field(default="")

    # Name of blob container in Azure Blob Storage [slimm test]
    BLOB_CONTAINER: str = Field(default="test")

    PERSISTENT_STORAGE_CONNECTION_STRING: str = Field(default="")

    # The configuration of this section is tightly coupled with the metadata that is
    # submitted with the results once they have been uploaded.

    # Four digit code indicating facility
    PLANT_CODE: str = Field(default="1210")

    # Name of the facility the robot is operating in
    PLANT_NAME: str = Field(default="Huldra")

    # Shortname of the facility the robot is operating in
    PLANT_SHORT_NAME: str = Field(default="HUA")

    # Name of robot
    ROBOT_NAME: str = Field(default="Placebot")

    # Unique identifier for this ISAR instance. Note that this should be a generated UUID.
    ISAR_ID: str = Field(default="00000000-0000-0000-0000-000000000000")

    # Serial number of the robot ISAR is connected to
    SERIAL_NUMBER: str = Field(default="0001")

    # Info about robot documentation
    DOCUMENTATION: List[DocumentInfo] = Field(default=[])

    # Coordinate reference system of facility
    COORDINATE_REFERENCE_SYSTEM: str = Field(default="EQUINOR:4100001")

    # Vertical reference system of facility
    VERTICAL_REFERENCE_SYSTEM: str = Field(default="MSL")

    # Rotational representations of reported results
    # Options [quaternion]
    MEDIA_ORIENTATION_REFERENCE_SYSTEM: str = Field(default="quaternion")

    # Timezone
    TIMEZONE: str = Field(default="UTC+00:00")

    # Data classification
    DATA_CLASSIFICATION: str = Field(default="internal")

    # List of MQTT Topics
    TOPIC_ISAR_STATUS: str = Field(default="status", validate_default=True)
    TOPIC_ISAR_MISSION: str = Field(default="mission", validate_default=True)
    TOPIC_ISAR_MISSION_ABORTED: str = Field(
        default="aborted_mission", validate_default=True
    )
    TOPIC_ISAR_TASK: str = Field(default="task", validate_default=True)
    TOPIC_ISAR_INSPECTION_RESULT: str = Field(
        default="inspection_result", validate_default=True
    )
    TOPIC_ISAR_INSPECTION_VALUE: str = Field(
        default="inspection_value", validate_default=True
    )
    TOPIC_ISAR_ROBOT_INFO: str = Field(default="robot_info", validate_default=True)
    TOPIC_ISAR_ROBOT_HEARTBEAT: str = Field(
        default="robot_heartbeat", validate_default=True
    )
    TOPIC_ISAR_STARTUP: str = Field(default="startup", validate_default=True)
    TOPIC_ISAR_INTERVENTION_NEEDED: str = Field(
        default="intervention_needed", validate_default=True
    )

    # List of MQTT Topics Expiry
    MQTT_ROBOT_HEARTBEAT_EXPIRY: int = Field(default=5)
    MQTT_TELEMETRY_EXPIRY: int = Field(default=10)
    MQTT_MISSION_AND_TASK_EXPIRY: int = Field(default=86400)

    # Logging

    #   Log handlers
    # Determines which log handlers are used by ISAR
    # Multiple log handlers can be chosen
    # Each handler will be called when logging
    # Selecting a different log handler than local may require certain access rights:
    #    - The Azure AI logger requires the 'APPLICATIONINSIGHTS_CONNECTION_STRING' to be set as an environment variable.
    # In general logging is configured in logging.conf
    LOG_HANDLER_LOCAL_ENABLED: bool = Field(default=True)
    LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED: bool = Field(default=False)

    DEBUG_LOG_FORMATTER: bool = Field(default=False)

    # You can set logger levels from environment variables ending with _LOG_LEVEL
    # For example, MQTT_LOG_LEVEL="DEBUG" will set the logger MQTT_LOGGER to DEBUG level
    # Handeled in log.py and only work for loggers defined in logging.conf

    REQUIRED_ROLE: str = Field(default="Mission.Control")

    @field_validator(
        "TOPIC_ISAR_STATUS",
        "TOPIC_ISAR_MISSION",
        "TOPIC_ISAR_TASK",
        "TOPIC_ISAR_ROBOT_INFO",
        "TOPIC_ISAR_ROBOT_HEARTBEAT",
        "TOPIC_ISAR_INSPECTION_RESULT",
        "TOPIC_ISAR_INSPECTION_VALUE",
        "TOPIC_ISAR_STARTUP",
        "TOPIC_ISAR_INTERVENTION_NEEDED",
        "TOPIC_ISAR_MISSION_ABORTED",
    )
    @classmethod
    def prefix_isar_topics(cls, v: Any, info: ValidationInfo) -> str:
        return f"isar/{info.data['ISAR_ID']}/{v}"

    model_config = SettingsConfigDict(
        env_prefix="ISAR_",
        env_file_encoding="utf-8",
        case_sensitive=True,
        extra="ignore",
    )


env = os.environ.get("ISAR_ENV")

if env == "test":
    load_dotenv(".env.test", override=True)
else:
    load_dotenv()
settings = Settings()


class RobotSettings(BaseSettings):
    def __init__(self) -> None:
        try:
            source = (
                files(f"{settings.ROBOT_PACKAGE}")
                .joinpath("config")
                .joinpath("settings.env")
            )
            with as_file(source) as eml:
                env_file = eml
        except ModuleNotFoundError:
            env_file = None
        super().__init__(_env_file=env_file)

    # ISAR steps the robot is capable of performing
    # This should be set in the robot package settings.env file
    CAPABILITIES: List[str] = Field(default=["take_image"])

    # Model of the robot which ISAR is connected to
    # This should be set in the robot package settings.env file
    ROBOT_MODEL: RobotModel = Field(default=RobotModel.Robot)  # type: ignore

    model_config = SettingsConfigDict(
        env_file_encoding="utf-8", case_sensitive=True, extra="ignore"
    )


robot_settings = RobotSettings()



from threading import Lock
from typing import Any, Dict, List, Optional

from robot_interface.telemetry.mqtt_client import MqttClientInterface


class MqttPublisherFake(MqttClientInterface):
    def __init__(self) -> None:
        self._lock = Lock()
        self.published: List[Dict[str, Any]] = []

    def publish(
        self,
        topic: str,
        payload: str,
        qos: int = 0,
        retain: bool = False,
        properties: Optional[Any] = None,
    ) -> None:
        with self._lock:
            self.published.append(
                {
                    "topic": topic,
                    "payload": payload,
                    "qos": qos,
                    "retain": retain,
                    "properties": properties,
                }
            )

    def last(self) -> Optional[Dict[str, Any]]:
        with self._lock:
            return self.published[-1] if self.published else None

    def count(self) -> int:
        with self._lock:
            return len(self.published)



import logging
from queue import Queue
from threading import Thread
from typing import Tuple

from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotRetrieveInspectionException,
)
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TASKS
from robot_interface.robot_interface import RobotInterface


class RobotUploadInspectionThread(Thread):
    def __init__(
        self,
        upload_queue: Queue,
        robot: RobotInterface,
        task: TASKS,
        mission: Mission,
    ):
        self.logger = logging.getLogger("robot")
        self.robot: RobotInterface = robot
        self.task: TASKS = task
        self.upload_queue = upload_queue
        self.mission: Mission = mission
        Thread.__init__(self, name=f"Robot inspection upload thread - {task.id}")

    def stop(self) -> None:
        return

    def run(self) -> None:
        try:
            inspection: Inspection = self.robot.get_inspection(task=self.task)
            if self.task.inspection_id != inspection.id:
                self.logger.warning(
                    f"The inspection_id of task ({self.task.inspection_id}) "
                    f"and result ({inspection.id}) is not matching. "
                    f"This may lead to confusions when accessing the inspection later"
                )

        except (RobotRetrieveInspectionException, RobotException) as e:
            error_message: ErrorMessage = ErrorMessage(
                error_reason=e.error_reason, error_description=e.error_description
            )
            self.task.error_message = error_message
            self.logger.error(
                f"Failed to retrieve inspections because: {e.error_description}"
            )
            return

        except Exception as e:
            self.logger.error(
                f"Failed to retrieve inspections because of unexpected error: {e}"
            )
            return

        if not inspection:
            self.logger.warning(
                f"No inspection result data retrieved for task {str(self.task.id)[:8]}"
            )

        inspection.metadata.tag_id = self.task.tag_id

        message: Tuple[Inspection, Mission] = (
            inspection,
            self.mission,
        )
        self.upload_queue.put(message)
        self.logger.info(
            f"Inspection result: {str(inspection.id)[:8]} queued for upload"
        )



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.home as Home
import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.returning_home as ReturningHome
import isar.state_machine.states.unknown_status as UnknownStatus
from isar.apis.models.models import MaintenanceResponse
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class InterventionNeeded(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[Maintenance.Maintenance]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        def release_intervention_needed_handler(
            should_release: EmptyMessage,
        ) -> Transition[UnknownStatus.UnknownStatus]:
            state_machine.events.api_requests.release_intervention_needed.response.trigger_event(
                EmptyMessage()
            )
            return UnknownStatus.transition()

        def _robot_status_event_handler(
            has_changed: EmptyMessage,
        ) -> Optional[Transition[Home.Home]]:
            robot_status: Optional[RobotStatus] = shared_state.robot_status.check()
            if robot_status == RobotStatus.Home:
                self.logger.info(
                    "Got robot status home while in intervention needed state. Leaving intervention needed state."
                )
                return Home.transition()
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[EmptyMessage](
                name="return_home_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: ReturningHome.transition_and_start_mission(True),
            ),
            EventHandlerMapping[EmptyMessage](
                name="release_intervention_needed_event",
                event=events.api_requests.release_intervention_needed.request,
                handler=release_intervention_needed_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="robot_status_event",
                event=events.robot_service_events.robot_status_changed,
                handler=_robot_status_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.InterventionNeeded,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[InterventionNeeded]:
    def _transition(state_machine: "StateMachine") -> InterventionNeeded:
        return InterventionNeeded(state_machine)

    return _transition



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.resuming_return_home import ResumingReturnHome
from isar.state_machine.states.return_home_paused import ReturnHomePaused


def test_transition_from_return_home_paused_to_resuming_return_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ReturnHomePaused(sync_state_machine)

    return_home_paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        return_home_paused_state.get_event_handler_by_name("resume_return_home_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ResumingReturnHome



from typing import TYPE_CHECKING, List

import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.paused as Paused
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Pausing(State):

    def __init__(self, state_machine: "StateMachine", mission_id: str):
        events = state_machine.events

        def _failed_pause_event_handler(
            error_message: ErrorMessage,
        ) -> Transition[Monitor.Monitor]:
            state_machine.logger.warning(
                f"Failed to pause mission: {error_message.error_description}"
            )
            return Monitor.transition_with_existing_mission(mission_id)

        def _successful_pause_event_handler(
            successful_pause: EmptyMessage,
        ) -> Transition[Paused.Paused]:
            return Paused.transition(mission_id)

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="failed_pause_event",
                event=events.robot_service_events.mission_failed_to_pause,
                handler=_failed_pause_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="successful_pause_event",
                event=events.robot_service_events.mission_successfully_paused,
                handler=_successful_pause_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Pausing,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition(mission_id: str) -> Transition[Pausing]:
    def _transition(state_machine: "StateMachine") -> Pausing:
        return Pausing(state_machine, mission_id=mission_id)

    return _transition



from typing import Optional, cast

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.going_to_recharging import GoingToRecharging
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.stopping_go_to_recharge import StoppingGoToRecharge


def test_stopping_to_recharge_goes_to_going_to_recharging(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = StoppingGoToRecharge(
        sync_state_machine, "mission_id"
    )
    stopping_go_to_recharge_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_go_to_recharge_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert not sync_state_machine.events.mqtt_queue.empty()

    mqtt_message = sync_state_machine.events.mqtt_queue.get(block=False)
    assert mqtt_message is not None
    mqtt_payload_topic = mqtt_message[0]
    assert mqtt_payload_topic is settings.TOPIC_ISAR_MISSION_ABORTED

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToRecharging


def test_return_home_goes_to_recharging_when_battery_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ReturningHome(sync_state_machine)
    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(10.0)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToRecharging



from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread
from typing import Callable, List, Optional

from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import InspectionTask
from robot_interface.models.robots.media import MediaConfig


class RobotInterface(metaclass=ABCMeta):
    """Interface to communicate with robots."""

    @abstractmethod
    def initiate_mission(self, mission: Mission) -> None:
        """Send a mission to the robot and initiate execution of the mission

        Parameters
        ----------
        mission: Mission

        Returns
        -------
        None

        Raises
        ------
        RobotAlreadyHomeException
            If the mission is a return home mission and the robot wish to disregard the
            mission as it is already at home
        RobotInfeasibleMissionException
            If the mission input is infeasible and the mission fails to be scheduled in
            a way that means attempting to schedule again is not necessary
        RobotException
            Will catch all RobotExceptions not previously listed and retry scheduling of
            the mission until the number of allowed retries is exceeded

        """
        raise NotImplementedError

    @abstractmethod
    def task_status(self, task_id: str) -> TaskStatus:
        """Gets the status of the currently active task on robot.

        Returns
        -------
        TaskStatus
            Status of the execution of current task.

        Raises
        ------
        RobotCommunicationTimeoutException or RobotCommunicationException
            If the robot package is unable to communicate with the robot API the fetching
            of task status will be attempted again until a certain number of retries
        RobotTaskStatusException
            If there was an error when retrieving the task status
        RobotException
            If the task status could not be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def mission_status(self, mission_id: str) -> MissionStatus:
        """Gets the status of the mission with ID mission_id on robot.

        Returns
        -------
        MissionStatus
            Status of the execution of mission.

        Raises
        ------
        RobotCommunicationTimeoutException or RobotCommunicationException
            If the robot package is unable to communicate with the robot API the fetching
            of mission status will be attempted again until a certain number of retries
        RobotMissionStatusException
            If there was an error when retrieving the mission status
        RobotException
            If the mission status could not be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the execution of the current task and corresponding mission.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to stop mission execution
            the action to stop will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to stop the mission

        """
        raise NotImplementedError

    @abstractmethod
    def pause(self) -> None:
        """Pauses the execution of the current task and stops the movement of the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to pause mission execution
            the action to pause will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to pause the mission

        """
        raise NotImplementedError

    @abstractmethod
    def resume(self) -> None:
        """Resumes the execution of the current task and continues the rest of the mission.

        Returns
        -------
        None

        Raises
        ------
        RobotActionException
            If the robot fails to perform the requested action to resume mission execution
            the action to resume will be attempted again until a certain number of retries
        RobotException
            Will catch other RobotExceptions and retry to resume the mission

        """
        raise NotImplementedError

    @abstractmethod
    def get_inspection(self, task: InspectionTask) -> Inspection:
        """Return the inspection connected to the given task.

        Parameters
        ----------
        task : InspectionTask

        Returns
        -------
        Inspection
            The inspection connected to the given task.
            get_inspection has responsibility to assign the inspection_id of the task
            to the inspection that it returns.

        Raises
        ------
        RobotRetrieveInspectionException
            If the robot package is unable to retrieve the inspections for the relevant
            mission or task an error message is logged and the state machine continues
        RobotException
            Catches other RobotExceptions that lead to the same result as a
            RobotRetrieveInspectionException

        """
        raise NotImplementedError

    @abstractmethod
    def register_inspection_callback(
        self, callback_function: Callable[[Inspection, Mission], None]
    ) -> None:
        """Register a function which should be run when inspection data is received
        asynchronously. This function should expect to receive an Inspection from.

        Parameters
        ----------
        callback_function : Callable[[Inspection, Mission], None]

        Returns
        -------
        None

        """
        raise NotImplementedError

    @abstractmethod
    def generate_media_config(self) -> Optional[MediaConfig]:
        """
        Generate a JSON containing the url and token needed to establish a media stream
        connection to a robot.

        Returns
        -------
        MediaConfig
            An object containing the connection information for a media stream connection

        """
        raise NotImplementedError

    @abstractmethod
    def get_telemetry_publishers(
        self, queue: Queue, isar_id: str, robot_name: str
    ) -> List[Thread]:
        """
        Set up telemetry publisher threads to publish regular updates for pose, battery
        level etc. from the robot to the MQTT broker. The publishers on the robot side
        will use the queue to pass messages to the MQTT Client on the ISAR side.

        The isar_id is passed to the robot to ensure the messages are published to the
        correct topics.

        Note that this functionality will only be utilized if MQTT is enabled in the
        settings.

        Returns
        -------
        List[Thread]
            List containing all threads that will be started to publish telemetry.

        """
        raise NotImplementedError

    @abstractmethod
    def robot_status(self) -> RobotStatus:
        """
        Method which returns an enum indicating the status of the robot.

        Returns
        -------
        RobotStatus
            Enum indicating if the robot may be reached by the isar-robot package.

        Raises
        -------
        RobotCommunicationException
            Raised if the robot package is unable to communicate with the robot API.
            The fetching of robot status will be attempted again until success
        RobotAPIException
            Raised if the robot package is able to communicate with the API but an error
            occurred while interpreting the response. The fetching of robot status will
            be attempted again until success
        RobotException
            Catches other RobotExceptions that may have occurred while retrieving the
            robot status. The fetching of robot status will be attempted again until
            success

        """
        raise NotImplementedError

    @abstractmethod
    def get_battery_level(self) -> float:
        """
        Method which returns the percent charge remaining in the robot battery.

        Returns
        -------
        float
            The battery percentage on the robot

        Raises
        -------
        RobotCommunicationException
            Raised if the robot package is unable to communicate with the robot API.
            The fetching of robot battery level will be attempted again until success
        RobotAPIException
            Raised if the robot package is able to communicate with the API but an error
            occurred while interpreting the response. The fetching of robot battery level
            will be attempted again until success
        RobotException
            Catches other RobotExceptions that may have occurred while retrieving the
            robot battery level. The fetching of robot battery level will
            be attempted again until success

        """
        raise NotImplementedError



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.resuming import Resuming


def test_transition_from_paused_to_resuming(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Paused(sync_state_machine, "mission_id")

    paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        paused_state.get_event_handler_by_name("resume_mission_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Resuming



from typing import Optional, cast

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.await_next_mission import AwaitNextMission
from isar.state_machine.states.going_to_lockdown import GoingToLockdown
from isar.state_machine.states.going_to_recharging import GoingToRecharging
from isar.state_machine.states.intervention_needed import InterventionNeeded
from isar.state_machine.states.return_home_paused import ReturnHomePaused
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.stopping_go_to_lockdown import StoppingGoToLockdown
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason


def test_transition_from_return_home_paused_to_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ReturnHomePaused(sync_state_machine)

    return_home_paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        return_home_paused_state.get_event_handler_by_name("send_to_lockdown_event")
    )

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)

    assert sync_state_machine.events.api_requests.send_to_lockdown.response.has_event()
    assert type(sync_state_machine.current_state) is GoingToLockdown

    going_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    lockdown_event_handler: Optional[EventHandlerMapping] = (
        going_to_lockdown_state.get_event_handler_by_name("mission_failed_to_resume")
    )
    assert lockdown_event_handler is not None

    transition = lockdown_event_handler.handler(
        ErrorMessage(
            error_description="Test going to lockdown resume return to home mission failed",
            error_reason=ErrorReason.RobotCommunicationException,
        )
    )

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is InterventionNeeded


def test_stopping_lockdown_transitions_to_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = StoppingGoToLockdown(
        sync_state_machine, "mission_id"
    )

    stopping_go_to_lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_go_to_lockdown_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert (
        sync_state_machine.events.api_requests.send_to_lockdown.response.check().lockdown_started
    )

    assert not sync_state_machine.events.mqtt_queue.empty()
    mqtt_message = sync_state_machine.events.mqtt_queue.get(block=False)
    assert mqtt_message is not None
    mqtt_payload_topic = mqtt_message[0]
    assert mqtt_payload_topic is settings.TOPIC_ISAR_MISSION_ABORTED

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToLockdown


def test_return_home_transitions_to_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("send_to_lockdown_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToLockdown


def test_recharging_transitions_to_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = GoingToRecharging(sync_state_machine)

    going_to_recharging_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_recharging_state.get_event_handler_by_name("send_to_lockdown_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToLockdown


def test_await_next_mission_transitions_to_going_to_lockdown(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = AwaitNextMission(sync_state_machine)

    await_next_mission_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        await_next_mission_state.get_event_handler_by_name("send_to_lockdown_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.api_requests.send_to_lockdown.response.check()
    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is GoingToLockdown



from pathlib import Path

src_dir = Path("source/api")

for file in src_dir.iterdir():
    print(f"Processing .rst-file: {file}")

    with open(file, "r") as f:
        lines = f.read()

    remove_strs = ["Submodules\n----------", "Subpackages\n-----------"]

    for remove_str in remove_strs:
        lines = lines.replace(remove_str, "")

    lines = lines.replace(" module\n=", "\n")
    lines = lines.replace(" package\n=", "\n")

    with open(file, "w") as f:
        f.write(lines)



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.going_to_lockdown as GoingToLockdown
import isar.state_machine.states.resuming_return_home as ResumingReturnHome
import isar.state_machine.states.returning_home as ReturningHome
import isar.state_machine.states.stopping_due_to_maintenance as StoppingDueToMaintenance
import isar.state_machine.states.stopping_paused_return_home as StoppingPausedReturnHome
from isar.apis.models.models import (
    ControlMissionResponse,
    LockdownResponse,
    MissionStartResponse,
)
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class ReturnHomePaused(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[ReturningHome.ReturningHome]]:
            if (
                battery_level is None
                or battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
            ):
                return None

            state_machine.events.state_machine_events.resume_mission.trigger_event(
                EmptyMessage()
            )
            return ReturningHome.transition_to_existing_mission()

        def _start_mission_event_handler(
            mission: Mission,
        ) -> Optional[Transition[StoppingPausedReturnHome.StoppingPausedReturnHome]]:
            if not state_machine.battery_level_is_above_mission_start_threshold():
                response = MissionStartResponse(
                    mission_id=None,
                    mission_started=False,
                    mission_not_started_reason="Robot battery too low",
                )
                state_machine.events.api_requests.start_mission.response.trigger_event(
                    response
                )
                return None
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingPausedReturnHome.transition(mission)

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[GoingToLockdown.GoingToLockdown]:
            events.api_requests.send_to_lockdown.response.trigger_event(
                LockdownResponse(lockdown_started=True)
            )
            state_machine.events.state_machine_events.resume_mission.trigger_event(
                EmptyMessage()
            )

            return GoingToLockdown.transition()

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[StoppingDueToMaintenance.StoppingDueToMaintenance]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to maintenance mode"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingDueToMaintenance.transition("")

        def _resume_mission_event_handler(
            should_resume: EmptyMessage,
        ) -> Transition[ResumingReturnHome.ResumingReturnHome]:
            state_machine.events.api_requests.resume_mission.response.trigger_event(
                ControlMissionResponse(success=True)
            )
            state_machine.events.state_machine_events.resume_mission.trigger_event(
                EmptyMessage()
            )
            return ResumingReturnHome.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[EmptyMessage](
                name="resume_return_home_event",
                event=events.api_requests.resume_mission.request,
                handler=_resume_mission_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[Mission](
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=_start_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.ReturnHomePaused,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[ReturnHomePaused]:
    def _transition(state_machine: "StateMachine") -> ReturnHomePaused:
        return ReturnHomePaused(state_machine)

    return _transition



from enum import Enum


class IsarStatus(Enum):
    Available = "available"
    ReturnHomePaused = "returnhomepaused"
    Paused = "paused"
    Busy = "busy"
    Home = "home"
    Offline = "offline"
    BlockedProtectiveStop = "blockedprotectivestop"
    ReturningHome = "returninghome"
    InterventionNeeded = "interventionneeded"
    Recharging = "recharging"
    Lockdown = "lockdown"
    GoingToLockdown = "goingtolockdown"
    GoingToRecharging = "goingtorecharging"
    Maintenance = "maintenance"
    Pausing = "pausing"
    PausingReturnHome = "pausingreturnhome"
    Stopping = "stopping"
    StoppingReturnHome = "stoppingreturnhome"



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.resuming as Resuming
import isar.state_machine.states.stopping_due_to_maintenance as StoppingDueToMaintenance
import isar.state_machine.states.stopping_go_to_lockdown as StoppingGoToLockdown
import isar.state_machine.states.stopping_paused_mission as StoppingPausedMission
from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Paused(State):

    def __init__(self, state_machine: "StateMachine", mission_id: str):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _stop_mission_event_handler(
            stop_mission_id: str,
        ) -> Optional[Transition[StoppingPausedMission.StoppingPausedMission]]:
            if mission_id == stop_mission_id or stop_mission_id == "":
                return StoppingPausedMission.transition_and_trigger_stop(
                    mission_id, True
                )
            else:
                state_machine.events.api_requests.stop_mission.response.trigger_event(
                    ControlMissionResponse(
                        success=False, failure_reason="Mission not found"
                    )
                )
                return None

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[StoppingPausedMission.StoppingPausedMission]]:
            if (
                battery_level is None
                or battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
            ):
                return None

            state_machine.publish_mission_aborted(
                mission_id, "Robot battery too low to continue mission", True
            )
            state_machine.logger.warning(
                "Cancelling current mission due to low battery"
            )
            return StoppingPausedMission.transition_and_trigger_stop(mission_id)

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[StoppingGoToLockdown.StoppingGoToLockdown]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to lockdown"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingGoToLockdown.transition(mission_id)

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[StoppingDueToMaintenance.StoppingDueToMaintenance]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to maintenance mode"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingDueToMaintenance.transition(mission_id)

        def _resume_mission_event_handler(
            should_resume: EmptyMessage,
        ) -> Optional[Transition[Resuming.Resuming]]:
            state_machine.events.api_requests.resume_mission.response.trigger_event(
                ControlMissionResponse(success=True)
            )
            state_machine.events.state_machine_events.resume_mission.trigger_event(
                EmptyMessage()
            )
            return Resuming.transition(mission_id)

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[str](
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=_stop_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="resume_mission_event",
                event=events.api_requests.resume_mission.request,
                handler=_resume_mission_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Paused,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition(mission_id: str) -> Transition[Paused]:
    def _transition(state_machine: "StateMachine") -> Paused:
        return Paused(state_machine, mission_id)

    return _transition



import logging
import time
from threading import Event, Thread

from isar.config.settings import settings
from isar.models.events import SharedState
from robot_interface.models.exceptions.robot_exceptions import RobotException
from robot_interface.robot_interface import RobotInterface


class RobotBatteryThread(Thread):
    def __init__(
        self,
        robot: RobotInterface,
        signal_thread_quitting: Event,
        shared_state: SharedState,
    ):
        self.logger = logging.getLogger("robot")
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.last_robot_battery_poll_time: float = time.time()
        self.force_battery_poll_next_iteration: bool = True
        Thread.__init__(self, name="Robot battery thread")

    def stop(self) -> None:
        return

    def _is_ready_to_poll_for_battery(self) -> bool:
        if self.force_battery_poll_next_iteration:
            self.force_battery_poll_next_iteration = False
            return True

        time_since_last_robot_battery_poll = (
            time.time() - self.last_robot_battery_poll_time
        )
        return (
            time_since_last_robot_battery_poll
            > settings.ROBOT_API_BATTERY_POLL_INTERVAL
        )

    def run(self) -> None:
        if self.signal_thread_quitting.is_set():
            return

        thread_check_interval = settings.THREAD_CHECK_INTERVAL

        while not self.signal_thread_quitting.wait(thread_check_interval):
            if not self._is_ready_to_poll_for_battery():
                continue
            try:
                self.last_robot_battery_poll_time = time.time()

                robot_battery_level = self.robot.get_battery_level()

                self.shared_state.robot_battery_level.update(robot_battery_level)
            except RobotException as e:
                self.logger.error(f"Failed to retrieve robot battery level: {e}")
                continue
            except Exception as e:
                self.logger.error(
                    f"Unhandled exception in robot battery service: {str(e)}"
                )
                continue
        self.logger.info("Exiting robot battery thread")



import logging

from fastapi import HTTPException
from opentelemetry import trace

from isar.apis.models.models import RobotInfoResponse
from isar.config.settings import robot_settings, settings
from isar.services.utilities.robot_utilities import RobotUtilities
from robot_interface.models.robots.media import MediaConfig

tracer = trace.get_tracer(__name__)


class RobotController:
    def __init__(
        self,
        robot_utilities: RobotUtilities,
    ):
        self.robot_utilities: RobotUtilities = robot_utilities
        self.logger = logging.getLogger("api")

    @tracer.start_as_current_span("generate_media_config")
    def generate_media_config(self) -> MediaConfig:
        media_config: MediaConfig = self.robot_utilities.generate_media_config()
        if media_config is None:
            raise HTTPException(
                status_code=204,
                detail="Robot has no media config",
            )
        return media_config

    @tracer.start_as_current_span("get_info")
    def get_info(self) -> RobotInfoResponse:
        return RobotInfoResponse(
            robot_package=settings.ROBOT_PACKAGE,
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            robot_capabilities=robot_settings.CAPABILITIES,
            robot_map_name=settings.DEFAULT_MAP,
            plant_short_name=settings.PLANT_SHORT_NAME,
        )



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.going_to_recharging import GoingToRecharging
from isar.state_machine.states.home import Home
from isar.state_machine.states.lockdown import Lockdown
from isar.state_machine.states.recharging import Recharging


def test_going_to_recharging_goes_to_recharge(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = GoingToRecharging(sync_state_machine)
    going_to_recharging_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        going_to_recharging_state.get_event_handler_by_name("mission_succeeded_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Recharging


def test_home_goes_to_recharging_when_battery_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Home(sync_state_machine)
    home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = home_state.get_event_handler_by_name(
        "robot_battery_update_event"
    )

    assert event_handler is not None

    transition = event_handler.handler(10.0)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Recharging


def test_lockdown_transitions_to_recharing_if_battery_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = Lockdown(sync_state_machine)

    lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        lockdown_state.get_event_handler_by_name("release_from_lockdown")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.api_requests.release_from_lockdown.response.check()
    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Recharging


def test_recharging_continues_when_battery_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Recharging(sync_state_machine)
    recharging_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        recharging_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(10.0)

    assert transition is None
